import datetime
from pathlib import Path

from ruamel.yaml import YAML

from lowball.models.authentication_models import Token, valid_token_id
from lowball.models.provider_models.auth_db import AuthDatabase
import re

class DefaultAuthDB(AuthDatabase):
    """Default Auth Database for Lowball Applications

    This class was developed to allow for a user to make full use of a Lowball
    application without having to create their own auth database provider. It makes
    use of the file system to store tokens.

    :param token_path: the base file system path at which tokens will be stored.
    :type token_path: str, Path
    """
    _YAML = YAML()

    def __init__(self, token_path=Path("/var/lib/lowball/authentication/tokens")):
        super(DefaultAuthDB, self).__init__()
        self.token_path = token_path

        try:
            self.token_path.mkdir(parents=True, exist_ok=True)
        except PermissionError:
            raise PermissionError("Lacking permissions to store tokens at configured route")

    @property
    def token_path(self):
        return self._token_path

    @token_path.setter
    def token_path(self, value):
        if not isinstance(value, (str, Path)):
            raise TypeError("token_path must be a string or a Path object")

        self._token_path = Path(value)

    def _dump_token(self, token_data):
        """Write a token to the file system.

        This will "store" a token on the system, i.e. writing token data to a
        flat file at the `token_path` location in YAML format.

        :param token_data: token metadata
        :type token_data: Token
        :return: None
        """
        token_id = token_data.token_id

        if not valid_token_id(token_id):
            raise ValueError(f"invalid token_id: {token_id}")

        token_file = open(self.token_path.joinpath(token_id), "w")

        self._YAML.dump(token_data.to_dict(), token_file)

    def _load_token(self, token_id):
        """Load token data from the file system into a token object.

        If there is a token file for the given token ID, but the data cannot be loaded
        into a Token object, then the file will automatically be deleted by this
        method, as this could cause issues with other operations.

        :param token_id: the identifier of the token to be loaded
        :return: token object
        :rtype: Token
        """
        if not valid_token_id(token_id):
            raise ValueError(f"invalid token_id: {token_id}")

        token_file = self.token_path.joinpath(token_id)

        if not token_file.exists():
            return None

        with open(token_file, "r") as f:
            data = self._YAML.load(f)

        try:
            return Token(**data)
        except (TypeError, ValueError):
            self._delete_token(token_id)

    def _delete_token(self, token_id):
        """Delete a token file from the file system.

        :param token_id: identifier of the token to delete
        :type token_id: str, Token
        :return: None
        """
        if not valid_token_id(token_id):
            raise ValueError(f"invalid token_id: {token_id}")

        token_file = self.token_path.joinpath(token_id)

        try:
            token_file.unlink()
        except FileNotFoundError:
            pass

    def add_token(self, token_object):
        """Add a token to the database."""
        self._dump_token(token_object)

    def lookup_token(self, token_id):
        """Lookup a token in the database."""
        return self._load_token(token_id)

    def revoke_token(self, token_id):
        """Remove a token from the database."""
        if isinstance(token_id, Token):
            token_id = token_id.token_id

        self._delete_token(token_id)

    def list_tokens(self):
        """List all the tokens in the database."""
        if not self.token_path.exists():
            return []

        tokens = []

        for token in self.token_path.iterdir():
            token = self._load_token(token)

            if token:
                tokens.append(token)

        return tokens

    def list_tokens_by_client_id(self, client_id):
        """List all the tokens in the database associated with a specific user."""
        tokens = self.list_tokens()

        return [token for token in tokens if token.client_id == client_id]

    def list_tokens_by_role(self, role):
        """List all the tokens in the database that have the specified role."""
        tokens = self.list_tokens()

        return [token for token in tokens if role in token.roles]

    def cleanup_tokens(self):
        """Remove all expired tokens from the database."""
        now = datetime.datetime.utcnow()
        tokens = self.list_tokens()

        expired = [token for token in tokens if now > token.expiration]

        for token in expired:
            self._delete_token(token.token_id)

    def revoke_all(self):
        """Remove all tokens from the database."""
        tokens = self.list_tokens()

        for token in tokens:
            self._delete_token(token.token_id)


__all__ = [
    "DefaultAuthDB"
]



import pytest

from lowball.builtins.providers.auth_provider import DefaultAuthProvider, DefaultAuthPackage
from lowball.models.authentication_models import ClientData
from lowball.exceptions import InvalidCredentialsException, MalformedAuthPackageException


class TestDefaultAuthProviderAuthPackage:
    @pytest.mark.parametrize("username, password", [
        (1, "userpassword"),
        ("user", 1)
    ])
    def test_init_raises_exception_when_bad_values_passed_in(self, username, password):
        with pytest.raises(TypeError):
            DefaultAuthPackage(username=username, password=password)

    def test_setting_username_to_non_string_raises_exception(self, default_auth_package, not_string):
        with pytest.raises(TypeError):
            default_auth_package.username = not_string

    def test_setting_password_to_non_string_raises_exception(self, default_auth_package, not_string):
        with pytest.raises(TypeError):
            default_auth_package.password = not_string


class TestDefaultAuthProvider:
    @pytest.mark.parametrize("username, password", [
        (1, "userpassword"),
        ("user", 1)
    ])
    def test_init_raises_exception_when_bad_values_passed_in(self, username, password):
        with pytest.raises(TypeError):
            DefaultAuthProvider(username=username, password=password)

    def test_setting_username_and_password_after_init_raises_exception(self):
        test_auth_provider = DefaultAuthProvider(username="admin", password="nimda")

        with pytest.raises(PermissionError):
            test_auth_provider.username = "somethingelse"

        with pytest.raises(PermissionError):
            test_auth_provider.password = "somethingelse"

    def test_auth_package_class_property_returns_expected_class(self, default_auth_provider):
        assert default_auth_provider.auth_package_class == DefaultAuthPackage

    def test_authenticate_returns_proper_data(self, default_auth_provider, default_auth_package):
        client_data = default_auth_provider.authenticate(default_auth_package)

        assert isinstance(client_data, ClientData)
        assert client_data.client_id == default_auth_provider.username
        assert client_data.roles == ["admin"]

    def test_passing_malformed_auth_package_to_authenticate_raises_exception(self, default_auth_provider,
                                                                             malformed_auth_package):
        with pytest.raises(MalformedAuthPackageException):
            default_auth_provider.authenticate(malformed_auth_package)

    def test_passing_invalid_auth_package_to_authenticate_raises_exception(self, default_auth_provider,
                                                                           invalid_auth_package):
        with pytest.raises(InvalidCredentialsException):
            default_auth_provider.authenticate(invalid_auth_package)

    def test_initialized_returns_true(self, default_auth_provider):
        assert default_auth_provider.initialized

    def test_get_client_returns_expected_client(self, default_auth_provider, get_client_expected_outcomes):

        client_id, expected_client_data = get_client_expected_outcomes

        client_data = default_auth_provider.get_client(client_id)

        if isinstance(expected_client_data, ClientData):
            assert isinstance(client_data, ClientData)
            assert client_data.to_dict() == expected_client_data.to_dict()
        else:
            assert client_data is None




import pytest

from flask import g, current_app

from lowball.authentication.wrappers import *
from lowball.exceptions import *


class TestValidateAuthHeaderFunction:
    # Failure case tests
    def test_passing_headers_without_auth_raises_exception(self, headers, request_no_headers):
        with pytest.raises(NoAuthHeaderException):
            validate_auth_header(headers)

    def test_passing_headers_with_invalid_auth_raises_exception(self, headers_invalid_auth, request_invalid_headers):
        with pytest.raises(InvalidAuthHeaderException):
            validate_auth_header(headers_invalid_auth)

    def test_passing_headers_with_expired_token_raises_exception(self, headers_with_token, request_token_expired):
        with pytest.raises(ExpiredTokenException):
            validate_auth_header(headers_with_token)

    def test_passing_headers_token_does_not_exist_in_database_raises_exception(self, headers_with_token,
                                                                               request_token_not_in_db):
        with pytest.raises(InvalidTokenException):
            validate_auth_header(headers_with_token)

    def test_passing_headers_token_data_does_not_match_database_token_raises_exception(self, headers_with_token,
                                                                                       request_unmatched_token):
        with pytest.raises(InvalidTokenException):
            validate_auth_header(headers_with_token)

    def test_making_request_with_no_authenticator_fails_gracefully(self, request_with_no_authdb, headers_with_token):
        with pytest.raises(NoAuthenticationDatabaseException):
            validate_auth_header(headers_with_token)

    # Success case tests
    def test_valid_auth_header_returns_proper_token_object(self, headers_with_token, request_valid_token, expected_token):
        decoded = validate_auth_header(headers_with_token)
        assert decoded == expected_token


class TestRequireAuthenticatedUserWrapper:
    def test_passing_invalid_tokens_raises_exception(self, request_invalid_headers):
        with pytest.raises(InvalidAuthHeaderException):
            @require_authenticated_user
            def test():
                pass

            test()

    def test_any_valid_token_is_accepted(self, request_any_role_token, expected_token):
        @require_authenticated_user
        def test():
            return True

        assert test()
        assert g.client_data == expected_token


class TestRequireAdminWrapper:
    def test_wrapper_raises_exception_when_no_admin_role_present(self, request_non_admin_token):
        with pytest.raises(InadequateRolesException):
            @require_admin
            def test():
                pass

            test()

    def test_wrapper_accepts_any_token_with_admin_role_included(self, request_admin_token, expected_token):
        @require_admin
        def test():
            return True

        assert test()
        assert g.client_data == expected_token


class TestRequireAnyOfTheseRolesWrapper:
    def test_wrapper_raises_exception_when_no_roles_present(self, request_token_no_roles):
        with pytest.raises(InadequateRolesException):
            @require_any_of_these_roles(roles=["admin", "lead", "analyst"])
            def test():
                pass

            test()

    def test_wrapper_works_with_any_combination_of_roles(self, request_admin_or_lead_or_analyst_token, expected_token):
        @require_any_of_these_roles(roles=["admin", "lead", "analyst"])
        def test():
            return True

        assert test()
        assert g.client_data == expected_token


class TestRequireAllOfTheseRolesWrapper:
    def test_wrapper_raises_exception_when_not_all_roles_present(self, request_not_all_roles_token):
        with pytest.raises(InadequateRolesException):
            @require_all_of_these_roles(roles=["admin", "lead", "analyst"])
            def test():
                pass

            test()

    def test_wrapper_accepts_token_with_all_roles(self, request_all_roles_token, expected_token):
        @require_all_of_these_roles(roles=["admin", "lead", "analyst"])
        def test():
            return True

        assert test()
        assert g.client_data == expected_token



import json
import pytest

from lowball.builtins.logging import DefaultFormatter


class TestDefaultFormatter:
    def test_passing_bad_values_to_init_raises_exception(self, not_string):
        with pytest.raises(TypeError):
            DefaultFormatter(date_format=not_string)

    @pytest.mark.parametrize("created_time, expected_datestr", [
        (1577858400.0, "2020-01-01 06:00:00.000000UTC"),
        (1609488121.000001, '2021-01-01 08:02:01.000001UTC')
    ])
    def test_format_time_method_returns_expected_date_strings(self, default_formatter, created_time, expected_datestr):
        formatted_time = default_formatter._format_time(created_time)
        assert formatted_time == expected_datestr
    
    def test_format_method_properly_formats_verbose_log(self, verbose_log, default_formatter,
                                                        expected_verbose_log_fields, expected_additional_field):
        formatted_log = default_formatter.format(verbose_log)
        assert isinstance(formatted_log, str)
        loaded_log = json.loads(formatted_log)
        assert all(field in loaded_log for field in expected_verbose_log_fields.keys())
        assert all(isinstance(loaded_log[field], data_type) for field, data_type in expected_verbose_log_fields.items())
        assert loaded_log["additional"] == expected_additional_field
        assert all(key not in loaded_log for key in ["requesting_client", "request_id", "client_token_id"])

    def test_format_method_properly_formats_default_log(self, default_log, default_formatter,
                                                        expected_default_log_fields, expected_additional_field):
        formatted_log = default_formatter.format(default_log)
        assert isinstance(formatted_log, str)
        loaded_log = json.loads(formatted_log)
        assert all(field in loaded_log for field in expected_default_log_fields.keys())
        assert all(isinstance(loaded_log[field], data_type) for field, data_type in expected_default_log_fields.items())
        assert loaded_log["additional"] == expected_additional_field
        assert all(key not in loaded_log for key in ["requesting_client", "request_id", "client_token_id"])

    def test_format_method_properly_formats_exception_log(self, exception_log, default_formatter,
                                                          expected_exception_log):
        formatted_log = default_formatter.format(exception_log)
        assert isinstance(formatted_log, str)
        loaded_log = json.loads(formatted_log)
        assert loaded_log == expected_exception_log
        assert all(key not in loaded_log for key in ["requesting_client", "request_id", "client_token_id"])
    
    def test_formatter_adds_flask_data_if_present_for_default_logs(self, default_formatter, default_log, mock_g,
                                                                   token_id, request_id):
        loaded_default_log = json.loads(default_formatter.format(default_log))
        assert loaded_default_log["requesting_client"] == "username"
        assert loaded_default_log["client_token_id"] == token_id
        assert loaded_default_log["request_id"] == request_id

    def test_formatter_adds_flask_data_if_present_for_verbose_logs(self, default_formatter, verbose_log, mock_g,
                                                                   token_id, request_id):
        loaded_verbose_log = json.loads(default_formatter.format(verbose_log))
        assert loaded_verbose_log["requesting_client"] == "username"
        assert loaded_verbose_log["client_token_id"] == token_id
        assert loaded_verbose_log["request_id"] == request_id

    def test_formatter_adds_flask_data_if_present_for_exception_logs(self, default_formatter, exception_log, mock_g,
                                                                     token_id, request_id):
        loaded_exception_log = json.loads(default_formatter.format(exception_log))
        assert loaded_exception_log["requesting_client"] == "username"
        assert loaded_exception_log["client_token_id"] == token_id
        assert loaded_exception_log["request_id"] == request_id



from pathlib import Path
from unittest.mock import call

import pytest
from ruamel.yaml import YAML

from lowball.builtins.providers import DefaultAuthDB


class TestAuthDBInit:
    def test_raise_clean_error_when_lacking_permissions_to_write_to_token_path(self, mkdir_raises_permission_error):
        with pytest.raises(PermissionError) as err:
            DefaultAuthDB()
        assert err.value.args[0] == "Lacking permissions to store tokens at configured route"


class TestAuthDBAttributes:
    def test_init_token_path_as_not_pathlike_object_raises_exception(self, not_pathlike):
        with pytest.raises(TypeError):
            DefaultAuthDB(token_path=not_pathlike)

    def test_token_path_attribute_access_returns_path_object(self, token_path, mocked_mkdir):
        test_auth_db = DefaultAuthDB(token_path=token_path)
        assert isinstance(test_auth_db.token_path, Path)

    def test_setting_token_path_to_not_pathlike_after_init_raises_exception(self, token_path, not_pathlike,
                                                                            mocked_mkdir):
        test_auth_db = DefaultAuthDB(token_path=token_path)

        with pytest.raises(TypeError):
            test_auth_db.token_path = not_pathlike

    def test_token_path_gets_created_on_init(self, token_path, mocked_mkdir):
        DefaultAuthDB(token_path=token_path)
        Path.mkdir.assert_called_once_with(parents=True, exist_ok=True)


class TestDumpTokenMethod:
    def test_dumping_tokens_with_path_like_ids_throws_errors(self, pathlike_token, token_path, mocked_mkdir):
        test_auth_db = DefaultAuthDB(token_path=token_path)

        with pytest.raises(ValueError):
            test_auth_db._dump_token(token_data=pathlike_token)

    def test_generating_token_creates_token_properly(self, token, mocked_yaml_dump, token_path, mock_open_file, mocked_mkdir):
        test_auth_db = DefaultAuthDB(token_path=token_path)
        test_auth_db._dump_token(token_data=token)

        opened_token_file = open(token_path)
        YAML.dump.assert_called_once_with(token.to_dict(), opened_token_file)


class TestLoadTokenMethod:
    def test_loading_tokens_with_path_like_ids_throws_errors(self, pathlike_token_id, token_path, mocked_mkdir):
        test_auth_db = DefaultAuthDB(token_path=token_path)

        with pytest.raises(ValueError):
            test_auth_db._load_token(token_id=pathlike_token_id)

    def test_load_non_existent_token_returns_none(self, token_path, token_id, mock_nonexistent_path, mocked_mkdir):
        test_auth_db = DefaultAuthDB(token_path=token_path)
        token_data = test_auth_db._load_token(token_id=token_id)

        assert token_data is None

    def test_loading_corrupted_token_file_returns_none_and_deletes_token(self, token_path, token_id,
                                                                         corrupted_token_file, mock_existent_path,
                                                                         mock_internal_revoke_token, mocked_mkdir):
        test_auth_db = DefaultAuthDB(token_path=token_path)
        token_data = test_auth_db._load_token(token_id=token_id)

        DefaultAuthDB._delete_token.assert_called_once_with(token_id)
        assert token_data is None

    def test_loading_token_data_returns_proper_data(self, token_path, token_id, token_file, token,
                                                    mock_existent_path, mocked_mkdir):
        test_auth_db = DefaultAuthDB(token_path=token_path)
        token_data = test_auth_db._load_token(token_id=token_id)

        assert token_data == token


class TestDeleteTokenMethod:
    def test_deleting_tokens_with_path_like_ids_throws_errors(self, pathlike_token_id, token_path, mocked_mkdir):
        test_auth_db = DefaultAuthDB(token_path=token_path)

        with pytest.raises(ValueError):
            test_auth_db._delete_token(token_id=pathlike_token_id)

    def test_deleting_non_existent_token_does_not_throw_exception(self, token_path, token_id, mocked_mkdir,
                                                                  mock_unlink_non_existent_file):
        test_auth_db = DefaultAuthDB(token_path=token_path)
        test_auth_db._delete_token(token_id=token_id)

        Path.unlink.assert_called_once()

    def test_deleting_valid_token_returns_true(self, token_path, token_id, mock_path_unlink, mocked_mkdir):
        test_auth_db = DefaultAuthDB(token_path=token_path)
        test_auth_db._delete_token(token_id=token_id)

        Path.unlink.assert_called_once()


class TestAuthDBMethods:
    def test_add_token_calls_internal_method_and_returns_true_for_success(self, default_auth_db, token,
                                                                          mock_default_auth_db_internals):
        default_auth_db.add_token(token)

        DefaultAuthDB._dump_token.assert_called_once_with(token)

    def test_lookup_token_returns_token_object(self, default_auth_db, token_id, mock_default_auth_db_internals,
                                               token):
        result = default_auth_db.lookup_token(token_id=token_id)

        DefaultAuthDB._load_token.assert_called_once_with(token_id)
        assert result == token

    def test_lookup_non_existent_token_returns_none(self, default_auth_db, token_id,
                                                    mock_internal_load_non_existent_token):
        result = default_auth_db.lookup_token(token_id=token_id)

        DefaultAuthDB._load_token.assert_called_once_with(token_id)
        assert result is None

    def test_revoke_token_calls_internal_method_and_returns_true(self, default_auth_db, token_identifier,
                                                                 mock_default_auth_db_internals, token_id):
        default_auth_db.revoke_token(token_id=token_identifier)

        DefaultAuthDB._delete_token.assert_called_once_with(token_id)

    def test_revoke_non_existent_token_returns_false(self, default_auth_db, token_identifier, token_id,
                                                     mock_internal_revoke_token):
        default_auth_db.revoke_token(token_id=token_identifier)

        DefaultAuthDB._delete_token.assert_called_once_with(token_id)

    def test_list_tokens_with_bad_token_path_returns_empty_list(self, default_auth_db, mock_nonexistent_path):
        result = default_auth_db.list_tokens()
        assert result == []

    def test_list_tokens_returns_expected_output(self, default_auth_db, mock_iterdir, mock_existent_path,
                                                 mock_load_multiple_tokens_internal, expected_list_tokens,
                                                 token_id1, token_id2, token_id3, token_id4, token_id5):
        calls = [
            call(default_auth_db.token_path.joinpath(token_id1)),
            call(default_auth_db.token_path.joinpath(token_id2)),
            call(default_auth_db.token_path.joinpath(token_id3)),
            call(default_auth_db.token_path.joinpath(token_id4)),
            call(default_auth_db.token_path.joinpath(token_id5)),
        ]
        token_list = default_auth_db.list_tokens()

        DefaultAuthDB._load_token.assert_has_calls(calls)
        assert token_list == expected_list_tokens

    def test_list_tokens_by_user_filters_tokens_properly(self, default_auth_db, mock_list_tokens,
                                                         expected_filter_by_user_result):
        result = default_auth_db.list_tokens_by_client_id(client_id="username")

        DefaultAuthDB.list_tokens.assert_called_once()
        assert result == expected_filter_by_user_result

    def test_list_tokens_by_roles_filters_tokens_properly(self, default_auth_db, mock_list_tokens,
                                                          expected_filter_by_role_result):
        result = default_auth_db.list_tokens_by_role(role="analyst")

        DefaultAuthDB.list_tokens.assert_called_once()
        assert result == expected_filter_by_role_result

    def test_cleanup_tokens_deletes_expired_tokens(self, default_auth_db, mock_list_tokens,
                                                   mock_default_auth_db_internals, token_id5, token_id4):
        calls = [
            call(token_id5),
            call(token_id4)
        ]

        default_auth_db.cleanup_tokens()

        DefaultAuthDB.list_tokens.assert_called_once()
        DefaultAuthDB._delete_token.assert_has_calls(calls)

    def test_revoke_all_tokens_makes_proper_calls(self, default_auth_db, mock_list_tokens,
                                                  mock_default_auth_db_internals,
                                                  token_id1, token_id2, token_id5, token_id3, token_id4):
        calls = [
            call(token_id1),
            call(token_id2),
            call(token_id5),
            call(token_id3),
            call(token_id4)
        ]

        default_auth_db.revoke_all()

        DefaultAuthDB.list_tokens.assert_called_once()
        DefaultAuthDB._delete_token.assert_has_calls(calls)



import json

from flask import Response

from lowball.builtins.response_class import LowballResponse


class TestLowballResponseClassForceType:
    def test_forces_type_properly_for_dicts(self, dict_return_value, client_with_response_class, expected_dict_return):
        forced_response = LowballResponse.force_type(dict_return_value)
        assert isinstance(forced_response, LowballResponse)
        assert forced_response.data == expected_dict_return

    def test_forces_type_properly_for_sequence_types(self, sequence_return_value, client_with_response_class):
        forced_response = LowballResponse.force_type(sequence_return_value)
        assert isinstance(forced_response, LowballResponse)
        assert forced_response.data == b'[1,2,3]\n'
    
    def test_forces_type_properly_for_ints(self, int_return_value):
        forced_response = LowballResponse.force_type(int_return_value)
        assert isinstance(forced_response, LowballResponse)
        assert forced_response.data == str(int_return_value).encode()
        
    def test_forces_type_properly_for_floats(self, float_return_value):
        forced_response = LowballResponse.force_type(float_return_value)
        assert isinstance(forced_response, LowballResponse)
        assert forced_response.data == str(float_return_value).encode()
        
    def test_forces_type_properly_for_complex_type(self, complex_return_value):
        forced_response = LowballResponse.force_type(complex_return_value)
        assert isinstance(forced_response, LowballResponse)
        assert forced_response.data == str(complex_return_value).encode()
    
    def test_forces_type_properly_for_memoryviews(self, memoryview_return_value):
        forced_response = LowballResponse.force_type(memoryview_return_value)
        assert isinstance(forced_response, LowballResponse)
        assert forced_response.data == bytes(memoryview_return_value)

    def test_passes_response_object_to_super(self, response_return_value, mocked_response_force_type):
        forced_response = LowballResponse.force_type(response_return_value)
        assert isinstance(forced_response, LowballResponse)
        Response.force_type.assert_called_once_with(response=response_return_value, environ=None)



from abc import ABC, abstractmethod


class AuthDatabase(ABC):
    """Base class for user-defined Auth Database classes.

    The data that is passed into the init is determined by the developer and the config that
    is passed into the init.
    """
    def __init__(self, **kwargs):
        pass

    @abstractmethod
    def add_token(self, token_object):
        """Add a token to the auth database.

        :param token_object: object containing all data to be written by auth db
        :type token_object: Token
        :return: None
        """
        pass

    @abstractmethod
    def lookup_token(self, token_id):
        """Lookup a token in the auth database

        :param token_id: identifier of the token in the auth database
        :type token_id: str
        :return: Token Object
        """
        pass

    @abstractmethod
    def revoke_token(self, token_id):
        """Revoke a token that is stored in the auth database.

        :param token_id: identifier of the token in the auth database
        :type token_id: str
        :return: None
        """
        pass

    @abstractmethod
    def list_tokens(self):
        """List all the tokens that are in the auth database.

        :return: List of Token Objects
        """
        pass

    @abstractmethod
    def list_tokens_by_client_id(self, client_id):
        """List all the tokens for a specific user.

        :param client_id: the username to lookup tokens for
        :type client_id: str
        :return: List of Token Objects
        """
        pass

    @abstractmethod
    def list_tokens_by_role(self, role):
        """List all tokens in the auth database that have a specific role.

        :param role: the role to lookup in the database
        :type role: str
        :return: List of Token Objects
        """
        pass

    @abstractmethod
    def cleanup_tokens(self):
        """Remove all expired tokens from the auth database.

        :return: None
        """
        pass

    @abstractmethod
    def revoke_all(self):
        """Revoke all tokens in the auth database.

        :return: None
        """
        pass



import pytest

from datetime import datetime
import jwt
from lowball.authentication.authentication import Authentication
from lowball.exceptions import InvalidAuthDataException, InvalidRequestingUserException, InvalidTokenLifeException, \
    InvalidTokenException, BadRequestException
from lowball.models.authentication_models import Token, ClientData


class TestAuthentication:
    def test_init_raises_type_error_when_invalid_arguments_passed_in(self):
        with pytest.raises(TypeError):
            Authentication(config="Not a Authentication Object")
    
    def test_setting_config_to_invalid_value_after_init_raises_type_error(self, auth_config, not_auth_config):
        test_auth = Authentication(config=auth_config)

        with pytest.raises(TypeError):
            test_auth.config = not_auth_config

    def test_create_token_creates_expected_tokens(self, token_create_rounds, auth):
        client_data, requesting_user, expiration, expected_expiration, expected_token_string = token_create_rounds

        token, token_data = auth.create_token(
            client_data.client_id,
            client_data.roles,
            requesting_client=requesting_user,
            expiration=expiration
        )
        assert token == expected_token_string
        assert isinstance(token_data, Token)
        assert token_data.roles == client_data.roles
        assert token_data.expiration == expected_expiration
        assert token_data.issued_by == client_data.client_id if requesting_user is None else requesting_user

    def test_expiration_greater_than_max_token_life_produces_value_error(self, client_data, requesting_user, auth,
                                                                         bad_expiration):
        with pytest.raises(InvalidTokenLifeException):
            auth.create_token(client_data.client_id, client_data.roles, requesting_client=requesting_user, expiration=bad_expiration)

    def test_decode_token_properly_decodes_token(self, auth, expected_token, expected_token_string):
        assert auth.decode_token(expected_token_string) == expected_token

    def test_validate_token_functions_as_expected(self, auth, expected_token_string, expired_token_string, invalid_expected_token_string):
        assert auth.validate_token(expected_token_string) == True
        assert auth.validate_token(expired_token_string) == False
        assert auth.validate_token(invalid_expected_token_string) == False

    def test_passing_bad_auth_data_into_create_token_produces_error(self, auth, not_auth_data, requesting_user,
                                                                    expiration):
        with pytest.raises(BadRequestException):
            auth.create_token(client_id=not_auth_data, roles=not_auth_data, requesting_client=requesting_user, expiration=expiration)

    def test_passing_bad_requesting_user_into_create_token_produces_error(self, auth, client_data, not_user,
                                                                          expiration):
        with pytest.raises(BadRequestException):
            auth.create_token(client_data.client_id, client_data.roles, requesting_client=not_user, expiration=expiration)

    def test_decode_token_raises_invalid_token_error_when_invalid_token_passed(self, auth, bad_token_rounds):

        with pytest.raises(InvalidTokenException):
            auth.decode_token(bad_token_rounds)


import datetime
from functools import wraps

from flask import request, current_app, g

from lowball.exceptions import NoAuthHeaderException, InvalidAuthHeaderException, ExpiredTokenException, \
    InvalidTokenException, InadequateRolesException, NoAuthenticationDatabaseException


def validate_auth_header(headers):
    """Validate and decode auth token in request headers.

    This helper function is used in each of the below wrappers, and is responsible to
    validate the format of the `Authorization` header where the Lowball token is
    supposed to reside.

    Requirements for successful validation:
    1. The current app must have a working auth database
    2. The `Authorization` header __must__ be present in the headers
    3. That header value __must__ be of the format `Bearer <token>`. The header value
       is split on the space character, and if the header value is properly formatted,
       this should result in a data structure that looks like ["Bearer", "<token>"]. If
       after splitting the header value on the space, the length of the resulting
       structure is not __exactly__ two, then the header is considered improperly formatted.
    4. The token must be able to be decoded by the `Authentication.decode_token` method
    5. The token cannot be expired.
    6. The token must match a token that is in the application authentication database __exactly__

    :param headers: Headers from request made to Lowball application
    :type headers: werkzeug.Headers
    :return: decoded token data
    :rtype: Token
    """
    if current_app.auth_db is None:
        raise NoAuthenticationDatabaseException

    if "Authorization" not in headers:
        raise NoAuthHeaderException

    auth_header = headers["Authorization"].split(" ")

    if len(auth_header) < 2 or auth_header[0] != "Bearer":
        raise InvalidAuthHeaderException

    token = auth_header[1]

    decoded = current_app.authenticator.decode_token(token)

    g.client_data = decoded

    if datetime.datetime.utcnow() > decoded.expiration:
        raise ExpiredTokenException

    database_token = current_app.auth_db.lookup_token(decoded.token_id)

    if database_token != decoded:
        raise InvalidTokenException

    return decoded


def require_authenticated_user(func):
    @wraps(func)
    def wrapped(*args, **kwargs):
        """Wrapper to ensure an authenticated user is accessing the route.

        This wrapper does not care about the roles that are associated with the user
        who is accessing the route. All that matters is that the token that is provided
        is in fact a valid token.
        """
        validate_auth_header(request.headers)

        return func(*args, **kwargs)

    return wrapped


def require_admin(func):
    @wraps(func)
    def wrapped(*args, **kwargs):
        """Wrapper to ensure a user with the admin role is accessing the route."""
        decoded = validate_auth_header(request.headers)

        if "admin" not in decoded.roles:
            raise InadequateRolesException

        return func(*args, **kwargs)

    return wrapped


def require_any_of_these_roles(roles=None):
    def decorator(func):
        @wraps(func)
        def wrapped(*args, **kwargs):
            """Wrapper to ensure that the user accessing the route has at least one of any number of roles."""
            r = [] if roles is None else roles

            decoded = validate_auth_header(request.headers)

            if not any(role in decoded.roles for role in r):
                raise InadequateRolesException

            return func(*args, **kwargs)

        return wrapped

    return decorator


def require_all_of_these_roles(roles=None):
    def decorator(func):
        @wraps(func)
        def wrapped(*args, **kwargs):
            """Wrapper to ensure that the user accessing the route has all of any number of roles."""
            r = [] if roles is None else roles

            decoded = validate_auth_header(request.headers)

            if not all(role in decoded.roles for role in r):
                raise InadequateRolesException

            return func(*args, **kwargs)

        return wrapped

    return decorator


__all__ = [
    "require_authenticated_user",
    "require_admin",
    "require_any_of_these_roles",
    "require_all_of_these_roles",
    "validate_auth_header"
]



import datetime
import json
from distutils.util import strtobool

from flask import g, current_app, request
from flask.blueprints import Blueprint

from lowball.authentication.wrappers import require_authenticated_user, require_admin
from lowball.exceptions import ListTokensInvalidReturnTypeException, ListTokensInvalidTokenInListException, \
    TokenNotFoundException, BadRequestException, InternalServerErrorException, NotFoundException, \
    MalformedProviderPackageException, InadequateRolesException, MalformedTokenIdException, \
    AuthenticationNotInitializedException
from lowball.models.authentication_models import Token, ClientData, valid_token_id

status = Blueprint("status", __name__, url_prefix="/builtins/status")

_JSON_HEADER = {"Content-Type": "application/json"}

@status.route("/", methods=["GET"], strict_slashes=False)
def get_application_status():
    """Get the status of the Lowball application.

    This view function returns data about the operation of the
    application, namely the name of the application, and whether
    the auth database and auth provider are initialized.

    The auth database is considered initialized if it is not None.

    We determine whether the auth provider is initialized by
    calling :meth:`initialized`. If an exception is thrown, we
    assume the provider is not initialized. If it returns something
    other than a bool, we consider it not initialized.

    :rtype: tuple
    :return: status information, status code, headers
    """
    app_name = current_app.lowball_config.meta.name

    try:
        auth_provider_initialized = current_app.auth_provider.initialized
    except:
        auth_provider_initialized = False

    if not isinstance(auth_provider_initialized, bool):
        auth_provider_initialized = False

    auth_db_initialized = False if current_app.auth_db is None else True

    status_information = {
        "name": app_name,
        "auth_provider_initialized": auth_provider_initialized,
        "auth_db_initialized": auth_db_initialized
    }

    return json.dumps(status_information), 200,  _JSON_HEADER



import builtins
import flask
import logging
import pytest

from logging.handlers import RotatingFileHandler
from pathlib import Path
from unittest.mock import Mock, mock_open

from lowball.builtins.logging import DefaultFormatter


@pytest.fixture
def log_record():
    return logging.makeLogRecord({
        "args": (),
        "asctime": "2020-12-21 00:00:00,000",
        "created": 1608530400.0,
        "filename": "test.py",
        "funcName": "function1",
        "levelname": "INFO",
        "levelno": 20,
        "lineno": 23,
        "module": "test",
        "msecs": 0,
        "msg": "Test Message",
        "name": "logger",
        "pathname": "/home/username/test.py",
        "process": 5985,
        "processName": "MainProcess",
        "stack_info": None,
        "thread": 139625468991296,
        "threadName": "MainThread",
        "another": "pair"
    })


@pytest.fixture
def date_format():
    return "%Y-%m-%d %H:%M:%S.%fUTC"


@pytest.fixture
def default_formatter(date_format):
    return DefaultFormatter(date_format=date_format)


@pytest.fixture(params=[
    ("DEBUG", 10),
])
def verbose_log(log_record, request):
    log_record.levelname = request.param[0]
    log_record.levelno = request.param[1]
    return log_record


@pytest.fixture(params=[
    ("INFO", 20),
    ("WARNING", 30),
])
def default_log(log_record, request):
    log_record.levelname = request.param[0]
    log_record.levelno = request.param[1]
    return log_record


@pytest.fixture
def exc_info():
    # Turns out it is really difficult to mock tracebacks
    # Code pulled from https://stackoverflow.com/a/19258720
    class FakeCode(object):
        def __init__(self, co_filename, co_name):
            self.co_filename = co_filename
            self.co_name = co_name

    class FakeFrame(object):
        def __init__(self, f_code, f_globals, f_back=None):
            self.f_code = f_code
            self.f_globals = f_globals
            self.f_back = f_back

    class FakeTraceback(object):
        def __init__(self, frames, line_nums):
            if len(frames) != len(line_nums):
                raise ValueError("Ya messed up!")
            self._frames = frames
            self._line_nums = line_nums
            self.tb_frame = frames[0]
            self.tb_lineno = line_nums[0]

        @property
        def tb_next(self):
            if len(self._frames) > 1:
                return FakeTraceback(self._frames[1:], self._line_nums[1:])

    class FakeException(Exception):
        def __init__(self, *args, **kwargs):
            self._tb = None
            super().__init__(*args, **kwargs)

        @property
        def __traceback__(self):
            return self._tb

        @__traceback__.setter
        def __traceback__(self, value):
            self._tb = value

        def with_traceback(self, value):
            self._tb = value
            return self

    # The fake code coincides with the filename and function that are in the
    # mocked log record defined above
    fake_code = FakeCode("test.py", "function1")

    # No need to fake globals here
    fake_frame = FakeFrame(fake_code, f_globals={})

    # Since the mocked log record above has the logging of the exception occurring
    # on line 23 of the fake `test.py` file, we will mock the line number that the
    # exception actually occurred on to be line 20
    exception_line_no = 20

    fake_tb = FakeTraceback(frames=[fake_frame], line_nums=[exception_line_no])
    return FakeException, FakeException("you messed up").with_traceback(fake_tb), fake_tb


@pytest.fixture(params=[
    ("ERROR", 10),
    ("CRITICAL", 50)
])
def exception_log(exc_info, log_record):
    log_record.levelname = "ERROR"
    log_record.levelno = 40
    log_record.exc_info = exc_info

    # Setting stack info to None. I believe that in the case that we are mocking,
    # it makes sense for it to be None
    log_record.stack_info = None
    return log_record


@pytest.fixture
def expected_additional_field():
    return {"another": "pair"}


@pytest.fixture
def expected_default_log_fields():
    return {
        "msg": str,
        "name": str,
        "args": list,
        "timestamp": str,
        "level": str,
        "additional": dict
    }


@pytest.fixture
def expected_verbose_log_fields(expected_default_log_fields):
    expected_default_log_fields.update({
        "filename": str,
        "func_name": str,
        "line_number": int,
        "module": str,
        "pathname": str,
        "process": int,
        "process_name": str,
        "thread": int,
        "thread_name": str
    })
    return expected_default_log_fields


@pytest.fixture
def expected_exception_log():
    return {
        "msg": "Test Message",
        "name": "logger",
        "args": [],
        "timestamp": "2020-12-21 06:00:00.000000UTC",
        "level": "ERROR",
        "additional": {"another": "pair"},
        "filename": "test.py",
        "func_name": "function1",
        "line_number": 23,
        "module": "test",
        "pathname": "/home/username/test.py",
        "process": 5985,
        "process_name": "MainProcess",
        "stack_info": None,
        "thread": 139625468991296,
        "thread_name": "MainThread",
        "exc_info": 'Traceback (most recent call last):\n  File "test.py", line 20, in function1\ntests.lowball.builtins.logging.conftest.exc_info.<locals>.FakeException: you messed up',
    }


@pytest.fixture(params=["/path/to/file.txt", Path("/path/to/file.txt")])
def filename(request):
    return request.param


@pytest.fixture(params=[
    logging.DEBUG,
    logging.INFO,
    logging.WARNING,
    logging.ERROR,
    logging.CRITICAL
])
def log_level(request):
    return request.param


@pytest.fixture(params=[2**20, 0])
def max_bytes(request):
    return request.param


@pytest.fixture(params=[0, 5, 20000])
def backup_count(request):
    return request.param


@pytest.fixture(params=[
    None,
    {},
    {"date_format": "%Y-%m-%d %H:%M:%S.%fUTC"}
])
def formatter(request):
    return request.param


@pytest.fixture
def monkeypatched_setters(monkeypatch):
    monkeypatch.setattr(RotatingFileHandler, "setFormatter", Mock(return_value=None))
    monkeypatch.setattr(RotatingFileHandler, "setLevel", Mock(return_value=None))


@pytest.fixture(params=[-1, 100, "not an int", {}, [], True, 1.1])
def bad_log_level(request):
    return request.param


@pytest.fixture
def bad_max_bytes():
    return -1


@pytest.fixture
def bad_backup_count():
    return -1


@pytest.fixture
def mock_open_file(monkeypatch):
    mock_file = mock_open()
    monkeypatch.setattr(builtins, "open", mock_file)
    return mock_file


@pytest.fixture
def request_id():
    return "01c19a54-1253-4b22-8a1a-004a511594d5"


@pytest.fixture
def mock_g(monkeypatch, expected_token, client, request_id):
    with client.test_client():
        with client.test_request_context() as client:
            flask.request.rid = request_id
            flask.g.client_data = expected_token
            yield client



from abc import ABC, abstractmethod

from lowball.exceptions import NotImplementedException


class AuthProvider(ABC):
    """Base class for user-defined Auth Provider classes.

    The data that is passed into the init is determined by the developer and the config that
    is passed into the init.
    """

    def __init__(self, **kwargs):
        pass

    @abstractmethod
    def authenticate(self, auth_package):
        """Authenticate using the provided auth package.

        :param auth_package: This should be defined by the developer
        :return: ClientData object
        """
        pass

    @property
    def initialized(self):
        """Property used to determine if the authentication provider is initialized.

        :return: whether this auth provider is initialized.
        :rtype: bool
        """
        return True

    @property
    @abstractmethod
    def auth_package_class(self):
        """Property used to return the AuthPackage class for this AuthProvider.

        :return: AuthPackage class accepted for this auth provider
        """
        pass

    @property
    def create_client_package_class(self):
        """Property used to return the CreateClientPackage class for this AuthProvider.

        :return: CreateClientPackage class accepted for this auth provider
        """
        raise NotImplementedException("create_client_package_class")

    @property
    def client_registration_package_class(self):
        """property used to return the ClientRegistrationPackage class for this auth provider

        :return: ClientRegistrationPackage class accepted for this auth provider

        """

        raise NotImplementedException("client_registration_package_class")

    @property
    def update_client_package_class(self):
        """Property used to return the UpdateClientPackage class for this AuthProvider.

        :return: AuthPackage class accepted for this auth provider
        """
        raise NotImplementedException("update_client_package_class")

    @property
    def self_update_client_package_class(self):
        """Property used to return the SelfUpdateClientPackage class for this AuthProvider.

        :return: AuthPackage class accepted for this auth provider
        """
        raise NotImplementedException("self_update_client_package_class")

    def get_client(self, client_id):
        """Get data for a user.



        :param client_id: user to get data for
        :type client_id: str
        :return: ClientData  or subclass
        """
        raise NotImplementedException("get_client")

    def create_client(self, client_registration_package):
        """Create a client.

        :type client_registration_package: ClientRegistrationPackage
        :param client_registration_package: user to create
        :return: dict or ClientData Object
        """
        raise NotImplementedException("create_client")

    def client_self_register(self, client_registration_package):
        """Allow a user to register themselves.

        This is intended to work the same as :meth:`create_user`, but to
        allow a user to perform the action on themself.

        :type client_registration_package: ClientRegistrationPackage
        :param client_registration_package: user to create
        :return: dict or ClientData Object
        """
        raise NotImplementedException("client_self_register")

    def enable_client(self, client_id):
        """Enable a client.

        :param client_id: client to enable
        :type client_id: str
        :return: dict/ClientData
        """
        raise NotImplementedException("enable_client")

    def disable_client(self, client_id):
        """Disable a user.

        :param client_id: user to disable
        :type client_id: str
        :return: dict
        """
        raise NotImplementedException("disable_client")

    def delete_client(self, client_id):
        """Delete a user.

        :param client_id: user to delete
        :type client_id: str
        :return: dict
        """
        raise NotImplementedException("delete_client")

    def add_roles(self, client_id, roles):
        """Add roles to a client_id

        :param client_id: user to add roles to
        :type client_id: str
        :param roles: list of roles to add to the user
        :type roles: list
        :return: dict/ClientData
        """
        raise NotImplementedException("add_roles")

    def update_client(self, update_client_package, client_id):
        """Update data for a client.

        :param update_client_package: client to update
        :type update_client_package: UpdateClientPackage
        :return: dict/ClientData
        """
        raise NotImplementedException("update_client")

    def client_self_update(self, self_update_client_package, client_id):
        """Allow a client to update data for their user.

        This is intended to work the same as :meth:`update_client`, but
        to allow a client to perform the action on himself.

        :type self_update_client_package: SelfUpdateClientPackage
        :param self_update_client_package: client info to update
        :type client_id: str
        :param client_id: client_id from token of authenticated client
        :return: dict
        """
        raise NotImplementedException("client_self_update")

    def list_clients(self):
        """List clients being tracked.

        :return: list(dict/ClientData)
        """
        raise NotImplementedException("list_clients")

    def delete_roles(self, client_id, roles):
        """Delete roles from a client.

        :param client_id: client to remove roles from
        :type client_id: str
        :param roles: list of roles to remove from the client
        :type roles: list
        :return: dict
        """
        raise NotImplementedException("delete_roles")


class AuthPackage(ABC):
    """Base class for Auth Package definitions.

    The data that is passed into the init is determined by the developer and the config that
    is passed into the init.
    """

    def __init__(self, **kwargs):
        pass


class CreateClientPackage(ABC):
    """Base Class for Client Create Package Definitions

    The data that is passed into init is determined by the developer and should expect a **kwargs style json pass

    """

    def __init__(self, **kwargs):
        pass


class ClientRegistrationPackage(ABC):
    """Base class for Client Registration Package Definitions

    """

    def __init__(self, **kwargs):
        pass


class UpdateClientPackage(ABC):
    """Base Class for Update Client Package Definitions
    """

    def __init__(self, **kwargs):
        pass


class SelfUpdateClientPackage(ABC):
    """Base Class for Self Update Client Package

    This class provides an alternative to the update Client Package to enable developers to differentiate
    what aspects of a client definition the client is able to change about themselves, in case it is different.
    """

    def __init__(self, **kwargs):
        pass



import datetime
from pathlib import Path
from unittest.mock import Mock, PropertyMock
from copy import deepcopy
import pytest
from flask import Flask, template_rendered, before_render_template, request_started, request_finished, \
    request_tearing_down, got_request_exception, appcontext_tearing_down, appcontext_popped, appcontext_pushed, \
    message_flashed
from flask.signals import Namespace
from werkzeug.datastructures import Headers
import jwt
from lowball import Lowball
from lowball.authentication.authentication import Authentication
from lowball.models.authentication_models import Token, ClientData
from lowball.models.config_models import Config, MetaConfig, AuthenticationConfig
from lowball.models.provider_models import AuthDatabase, AuthProvider
from lowball.models.provider_models.auth_provider import (
    AuthPackage,
    ClientRegistrationPackage,
    UpdateClientPackage,
    SelfUpdateClientPackage,
    CreateClientPackage
)

FAKE_NOW = datetime.datetime(2020, 1, 1, 0, 0, 0, 0)


@pytest.fixture
def fake_now():
    return FAKE_NOW

@pytest.fixture
def expired_time():
    return datetime.datetime(2019, 1, 1, 0, 0, 0, 0)

@pytest.fixture
def token_secret():
    return "supersecrettokensecret"

@pytest.fixture
def bad_token_secret():
    return "notactuallybadbutdifferent"

@pytest.fixture(params=[1, True, {}, object()])
def not_datetime(request):
    return request.param


@pytest.fixture(params=[{}, [], object(), 0, 1.1, True])
def not_string(request):
    return request.param


@pytest.fixture
def token_id():
    return "thesimpletokenid"


@pytest.fixture
def expected_token(token_id):
    return Token(
        cid="username",
        r=[],
        cts=datetime.datetime(2019, 11, 1),
        ets=datetime.datetime(2020, 12, 1),
        rcid="another_user",
        tid=token_id
    )

@pytest.fixture
def expired_token(token_id):
    return Token(
        cid="username",
        r=[],
        cts=datetime.datetime(2019, 11, 1),
        ets=datetime.datetime(2019, 12, 1),
        rcid="another_user",
        tid=token_id
    )

@pytest.fixture
def expected_token_string(expected_token, token_secret):

    return jwt.encode(expected_token.to_dict(), token_secret, algorithm="HS256")

@pytest.fixture
def expired_token_string(expired_token, token_secret):
    return jwt.encode(expired_token.to_dict(), token_secret, algorithm="HS256")

@pytest.fixture
def invalid_expected_token_string(expected_token, bad_token_secret):
    return jwt.encode(expected_token.to_dict(), bad_token_secret, algorithm="HS256")


@pytest.fixture
def auth_config(token_secret):
    return AuthenticationConfig(
        default_token_life=1800,
        max_token_life=31536000,
        token_secret=token_secret
    )


@pytest.fixture
def mock_auth_db():
    class MockAuthDB(AuthDatabase):
        def add_token(self, token_object):
            pass

        def lookup_token(self, token_id):
            pass  # TODO this is what needs to be mocked, a couple different cases

        def revoke_token(self, token_id):
            pass

        def list_tokens(self):
            pass

        def list_tokens_by_client_id(self, client_id):
            pass

        def list_tokens_by_role(self, role):
            pass

        def cleanup_tokens(self):
            pass

        def revoke_all(self):
            pass

    return MockAuthDB()


@pytest.fixture
def auth_package_test_class():
    class TestAuthPackage(AuthPackage):

        def __init__(self, client_id, client_secret, **kwargs):
            self.client_id = client_id
            self.client_secret = client_secret
            AuthPackage.__init__(self)

    return TestAuthPackage


@pytest.fixture
def client_registration_package_test_class():
    class TestClientRegistrationPackage(ClientRegistrationPackage):

        def __init__(self, client_id, client_info, **kwargs):
            self.client_id = client_id
            self.client_info = client_info
            ClientRegistrationPackage.__init__(self)

    return TestClientRegistrationPackage

@pytest.fixture
def client_update_package_test_calss():
    class TestClientUpdatePackage(UpdateClientPackage):

        def __init__(self, client_info1, client_info2, **kwargs):
            self.client_info1 = client_info1
            self.client_info2 = client_info2
            UpdateClientPackage.__init__(self)

    return TestClientUpdatePackage

@pytest.fixture
def client_self_update_package_test_calss():
    class TestClientSelfUpdatePackage(SelfUpdateClientPackage):

        def __init__(self, client_info1, client_info2, **kwargs):
            self.client_info1 = client_info1
            self.client_info2 = client_info2
            SelfUpdateClientPackage.__init__(self)

    return TestClientSelfUpdatePackage


@pytest.fixture
def mock_auth_provider_base(auth_package_test_class, client_data):
    class MockAuthProvider(AuthProvider):

        @property
        def auth_package_class(self):
            return auth_package_test_class

        def authenticate(self, auth_package):

            return client_data

    return MockAuthProvider()


@pytest.fixture
def auth(auth_config):
    return Authentication(config=auth_config)


@pytest.fixture(autouse=True)
def client(mock_auth_db, auth):
    test_app = Flask("test")
    test_app.config["TESTING"] = True
    test_app.auth_db = mock_auth_db
    test_app.authenticator = auth
    return test_app


@pytest.fixture(params=[1, [1, 2, 3], ("tuple",), {1, 2, 3}, {}])
def not_pathlike(request):
    return request.param


@pytest.fixture(autouse=True)
def patched_utcnow(monkeypatch):
    class MockedDatetime(datetime.datetime):
        @classmethod
        def utcnow(cls):
            return FAKE_NOW

    monkeypatch.setattr(datetime, "datetime", MockedDatetime)


@pytest.fixture(params=[1800, "1800"])
def default_token_life(request):
    return request.param


@pytest.fixture(params=[1800, "1800"])
def max_token_life(request):
    return request.param


@pytest.fixture
def authentication_config(default_token_life, max_token_life, token_secret):
    return AuthenticationConfig(
        default_token_life=default_token_life,
        max_token_life=max_token_life,
        token_secret=token_secret
    )


@pytest.fixture
def meta_config():
    return MetaConfig(name="APP_NAME", base_route="/base/route", description="description", tags=["tag", "list"])


@pytest.fixture
def lowball_config(meta_config, authentication_config):
    logging_config = {
        "filename": "lowball.log",
        "formatter": {
            "date_format": "%Y-%m-%d %H:%M:%S.%fUTC"
        }
    }
    return Config(meta=meta_config, authentication=authentication_config, logging=logging_config)


@pytest.fixture
def mocked_lowball_init(monkeypatch):
    monkeypatch.setattr(Flask, "register_blueprint", Mock())
    monkeypatch.setattr(Flask, "register_error_handler", Mock())
    monkeypatch.setattr(Lowball, "register_request_finished_handler", Mock())


@pytest.fixture
def mocked_mkdir(monkeypatch):
    monkeypatch.setattr(Path, "mkdir", Mock())


@pytest.fixture
def lowball_app(lowball_config, mocked_mkdir):
    return Lowball(config=lowball_config)


@pytest.fixture
def signal_subscriber():
    def subscriber(sender, **extra):
        pass

    return subscriber


@pytest.fixture
def other_signal_subscriber():
    def other_subscriber(sender, **extra):
        pass

    return other_subscriber


@pytest.fixture
def mocked_signals(monkeypatch):
    monkeypatch.setattr(template_rendered, "connect", Mock())
    monkeypatch.setattr(before_render_template, "connect", Mock())
    monkeypatch.setattr(request_started, "connect", Mock())
    monkeypatch.setattr(request_finished, "connect", Mock())
    monkeypatch.setattr(got_request_exception, "connect", Mock())
    monkeypatch.setattr(request_tearing_down, "connect", Mock())
    monkeypatch.setattr(appcontext_tearing_down, "connect", Mock())
    monkeypatch.setattr(appcontext_pushed, "connect", Mock())
    monkeypatch.setattr(appcontext_popped, "connect", Mock())
    monkeypatch.setattr(message_flashed, "connect", Mock())


@pytest.fixture
def custom_signal():
    ns = Namespace()
    custom_sig = ns.signal("custom_signal")
    custom_sig.connect = Mock()

    return custom_sig

@pytest.fixture
def headers():
    return Headers()


@pytest.fixture
def headers_with_token(headers, jwt_token):
    headers.add("Authorization", f"Bearer {jwt_token}")
    return headers


@pytest.fixture(params=[
    "Nothing about this is correct",
    "eyJ0eXAiOiJKV1QiLCJhbGciOiJIUzI1NiJ9.eyJ2ZXJzaW9uIjoiMSIsInVzZXJuYW1lIjoidXNlcm5hbWUxIiwicm9sZXMiOlsiYWRtaW4iLCJsZWFkIiwiYW5hbHlzdCJdLCJjcmVhdGVkIjoiMjAyMC0wMS0wMSAwMDowMDowMCIsImV4cGlyYXRpb24iOiIyMDIwLTAxLTAxIDAwOjMwOjAwIiwiaXNzdWVkX2J5IjoidXNlcm5hbWUxIiwidG9rZW5faWQiOiI0NmY5NjQzNS04N2ZhLTQzNjMtYWM1Ny1lZmRjNzI2NzhiMzcifQ.Dj92CngpXF9HZU-czed2qqyOocHnyccTzNt5VKIQN6U",
])
def headers_invalid_auth(request):
    headers = Headers()
    headers.add("Authorization", request.param)
    return headers


@pytest.fixture
def token(token_id, all_roles):
    return Token(
        cid="username",
        r=all_roles,
        cts=datetime.datetime(2020, 1, 1),
        ets=datetime.datetime(2020, 2, 1),
        rcid="username",
        tid=token_id
    )

@pytest.fixture
def jwt_token(token, token_secret):
    return jwt.encode(token.to_dict(), token_secret, algorithm="HS256")

@pytest.fixture
def username():
    return "username"

@pytest.fixture
def admin_role():
    return ["admin"]

@pytest.fixture
def base_role1():
    return ["role1"]

@pytest.fixture
def base_role2():
    return ["role2"]

@pytest.fixture
def base_roles(base_role1, base_role2):
    return base_role1 + base_role2

@pytest.fixture
def all_roles(admin_role, base_roles):
    return admin_role + base_roles

@pytest.fixture(params=[1, 1.1, {}, True, object(), "string"])
def not_list(request):
    return request.param


# TOKEN IDS
@pytest.fixture
def token_id1():
    return "abcdef0123456789"

@pytest.fixture
def token_id2():
    return "bbcdef0123456789"

@pytest.fixture
def token_id3():
    return "cbcdef0123456789"

@pytest.fixture
def token_id4():
    return "dbcdef0123456789"

@pytest.fixture
def token_id5():
    return "ebcdef0123456789"

@pytest.fixture
def token_id6():
    return "fbcdef0123456789"

@pytest.fixture
def token_id7():
    return "gbcdef0123456789"

@pytest.fixture
def token_id8():
    return "hbcdef0123456789"

@pytest.fixture
def token_id9():
    return "ibcdef0123456789"

@pytest.fixture
def token_id10():
    return "jbcdef0123456789"

@pytest.fixture
def token_id11():
    return "kbcdef0123456789"

@pytest.fixture
def token_id12():
    return "lbcdef0123456789"


# CLIENT IDS
@pytest.fixture
def client_id_admin1():
    return "admin1"

@pytest.fixture
def client_id_admin2():
    return "admin2"

@pytest.fixture
def client_id_normal1():
    return "non_admin1"

@pytest.fixture
def client_id_normal2():
    return "non_admin2"

# TOKENS
@pytest.fixture
def token1_admin1(token_id1, client_id_admin1, token, admin_role):
    modified_token = deepcopy(token)
    modified_token.token_id = token_id1
    modified_token.client_id = client_id_admin1
    modified_token.roles = admin_role
    return modified_token

@pytest.fixture
def token2_admin1(token_id2, client_id_admin1, token, all_roles):
    modified_token = deepcopy(token)
    modified_token.token_id = token_id2
    modified_token.client_id = client_id_admin1
    modified_token.roles = all_roles
    return modified_token

@pytest.fixture
def token3_admin1_expired(token_id3, client_id_admin1, token, admin_role, expired_time):
    modified_token = deepcopy(token)
    modified_token.token_id = token_id3
    modified_token.client_id = client_id_admin1
    modified_token.roles = admin_role
    modified_token.expiration = expired_time
    return modified_token

@pytest.fixture
def token1_admin2(token_id4, client_id_admin2, token, admin_role):
    modified_token = deepcopy(token)
    modified_token.token_id = token_id4
    modified_token.client_id = client_id_admin2
    modified_token.roles = admin_role
    return modified_token

@pytest.fixture
def token2_admin2(token_id5, client_id_admin2, token, all_roles):
    modified_token = deepcopy(token)
    modified_token.token_id = token_id5
    modified_token.client_id = client_id_admin2
    modified_token.roles = all_roles
    return modified_token

@pytest.fixture
def token3_admin2_expired(token_id6, client_id_admin2, token, admin_role, expired_time):
    modified_token = deepcopy(token)
    modified_token.token_id = token_id6
    modified_token.client_id = client_id_admin2
    modified_token.roles = admin_role
    modified_token.expiration = expired_time
    return modified_token

@pytest.fixture
def token1_normal1(token_id7, client_id_normal1, token, base_role1):
    modified_token = deepcopy(token)
    modified_token.token_id = token_id7
    modified_token.client_id = client_id_normal1
    modified_token.roles = base_role1
    return modified_token

@pytest.fixture
def token2_normal1(token_id8, client_id_normal1, token, base_roles):
    modified_token = deepcopy(token)
    modified_token.token_id = token_id8
    modified_token.client_id = client_id_normal1
    modified_token.roles = base_roles
    return modified_token

@pytest.fixture
def token3_normal1_expired(token_id9, client_id_normal1, token, base_role2, expired_time):
    modified_token = deepcopy(token)
    modified_token.token_id = token_id9
    modified_token.client_id = client_id_normal1
    modified_token.roles = base_role2
    modified_token.expiration = expired_time
    return modified_token

@pytest.fixture
def token1_normal2(token_id10, client_id_normal2, token, base_role2):
    modified_token = deepcopy(token)
    modified_token.token_id = token_id10
    modified_token.client_id = client_id_normal2
    modified_token.roles = base_role2
    return modified_token

@pytest.fixture
def token2_normal2(token_id11, client_id_normal2, token, base_roles):
    modified_token = deepcopy(token)
    modified_token.token_id = token_id11
    modified_token.client_id = client_id_normal2
    modified_token.roles = base_roles
    return modified_token

@pytest.fixture
def token3_normal2_expired(token_id12, client_id_normal2, token, base_role1, expired_time):
    modified_token = deepcopy(token)
    modified_token.token_id = token_id12
    modified_token.client_id = client_id_normal2
    modified_token.roles = base_role1
    modified_token.expiration = expired_time
    return modified_token


@pytest.fixture
def client_data(all_roles):
    return ClientData("username1", roles=all_roles)



from logging import INFO, ERROR
from unittest.mock import call

from flask import current_app, request, g

from lowball.builtins.signal_subscribers import request_finished_log


class TestRequestFinishedLogSubscriber:
    def test_does_not_do_anything_if_there_is_no_request_environment(self, context_no_request_env, base_response):
        request_finished_log(current_app, base_response)
        current_app.logger._log.assert_not_called()

    def test_response_is_not_exception_handled_properly(self, context_normal_env, base_response):
        log_dict = {
            "user_agent": request.environ.get("HTTP_USER_AGENT", "None"),
            "src_ip": request.environ.get("HTTP_X_FORWARDED_FOR", request.environ.get("REMOTE_ADDR")),
            "http_method": request.method,
            "url": request.environ["werkzeug.request"].full_path,
            "status_code": base_response.status_code,
            "client_data": {
                "requesting_client": g.client_data.client_id,
                "client_token_id": g.client_data.token_id
            }
        }

        log_call = call(
            INFO,                              # Logging level
            {"result": base_response.status},  # The msg of the _log call
            (),                                # container for *args
            extra=log_dict                     # "extra" kwarg
        )
        request_finished_log(current_app, base_response)
        current_app.logger._log.assert_has_calls([log_call])

    def test_response_is_exception_handled_properly(self, context_exception_env, exception_response):
        log_dict = {
            "user_agent": request.environ.get("HTTP_USER_AGENT", "None"),
            "src_ip": request.environ.get("HTTP_X_FORWARDED_FOR", request.environ.get("REMOTE_ADDR")),
            "http_method": request.method,
            "url": request.environ["werkzeug.request"].full_path,
            "status_code": exception_response.status_code,
            "client_data": {
                "requesting_client": g.client_data.client_id,
                "client_token_id": g.client_data.token_id
            }
        }

        msg = {"result": exception_response.status, "error_information": g.response_exception_log_data}

        log_call = call(
            ERROR,          # Logging level
            msg,            # The msg of the _log call
            (),             # container for *args
            extra=log_dict  # "extra" kwarg
        )
        request_finished_log(current_app, exception_response)
        current_app.logger._log.assert_has_calls([log_call])



import datetime
from unittest.mock import call

from flask import request

from lowball.exceptions import InadequateRolesException
from lowball.models.authentication_models import Token

class TestLogin:
    ROUTE = "/builtins/auth"

    # test no auth provider
    def test_user_login_attempt_fails_with_no_auth_provider(
            self, lowball_app_no_auth_provider, expected_no_auth_provider_error_response):
        response = lowball_app_no_auth_provider.post(self.ROUTE)
        assert response.status_code == 503
        assert response.json["message"] == expected_no_auth_provider_error_response

    # test no auth db
    def test_user_login_attempt_fails_with_no_auth_database(self, lowball_app_no_db, expected_no_auth_db_response):
        response = lowball_app_no_db.post(self.ROUTE)
        assert response.status_code == 503
        assert response.json["message"] == expected_no_auth_db_response

    # test auth provider no auth package class
    def test_user_login_attempt_fails_with_no_defined_auth_package_class(self,
                                                                         lowball_app_mock_providers):
        response = lowball_app_mock_providers.post(self.ROUTE, json={"some": "json"})

        assert response.status_code == 501
        assert response.json["message"] == "501 Not Implemented: auth_package_class not implemented"

    # test bad json input
    def test_user_login_attempt_fails_with_incorrect_auth_package_input(self,
                                                                        lowball_app_mock_providers,
                                                                        does_not_match_auth_package_request,
                                                                        mock_auth_package_class):
        response = lowball_app_mock_providers.post(self.ROUTE, json=does_not_match_auth_package_request)
        print(response.json)
        assert response.status_code == 400
        assert "The Authentication Request Did Not Supply The Required Data" in response.json["message"]

    # test success - calls authenticate and gets token data - calls authdb add token, returns data
    def test_user_login_attempt_success_flow(self,
                                             lowball_app_mock_providers,
                                             mock_auth_package_class,
                                             mock_authenticate_basic,
                                             success_request,
                                             mock_create_token,
                                             token,
                                             jwt_token,
                                             mock_add_token):
        response = lowball_app_mock_providers.post(self.ROUTE, json=success_request)
        assert response.status_code == 200
        assert "token" in response.json
        assert "token_data" in response.json
        assert response.json["token"] == jwt_token
        assert response.json["token_data"] == token.to_dict()
        lowball_app_mock_providers.application.auth_db.add_token.assert_called_once_with(token)
        lowball_app_mock_providers.application.authenticator.create_token.assert_called_once_with(client_id=token.client_id,
                                                                                                  roles=token.roles)


class TestLogout:
    ROUTE = "/builtins/auth"

    # test no auth db
    def test_fails_with_no_auth_database(self, lowball_app_no_db, expected_no_auth_db_response, basic_header):
        response = lowball_app_no_db.delete(self.ROUTE, headers=basic_header)
        assert response.status_code == 503
        assert response.json["message"] == expected_no_auth_db_response

    # assert not authenticated fails
    def test_self_update_attempt_fails_when_not_authenticated_or_token_expired(self, lowball_app_mock_providers,
                                                                               not_authenticated_or_token_expired_header):
        response = lowball_app_mock_providers.delete(self.ROUTE, headers=not_authenticated_or_token_expired_header,
                                                   json={})
        assert response.status_code != 200
        assert any([
            "Token Has Expired" in response.json["message"],
            "No token provided" in response.json["message"],
            "Token Is Invalid" in response.json["message"]
        ])

    # test no auth db
    def test_user_logout_attempt_fails_with_no_auth_database(self, lowball_app_no_db, expected_no_auth_db_response,
                                                             basic_header):
        response = lowball_app_no_db.delete(self.ROUTE, headers=basic_header)
        assert response.status_code == 503
        assert response.json["message"] == expected_no_auth_db_response

    # test calls revoke token with token id, returns 204 no data
    def test_user_logout_success_flow(self, lowball_app_mock_providers, basic_header, token,
                                      mock_decode_token, mock_lookup_token, mock_revoke_token):
        response = lowball_app_mock_providers.delete(self.ROUTE, headers=basic_header)
        assert response.status_code == 204
        assert not response.get_data(as_text=True)
        lowball_app_mock_providers.application.auth_db.revoke_token.assert_called_once_with(token.token_id)


class TestWhoami:
    ROUTE = "/builtins/auth"

    # test no auth db
    def test_fails_with_no_auth_database(self, lowball_app_no_db, expected_no_auth_db_response, basic_header):
        response = lowball_app_no_db.get(self.ROUTE, headers=basic_header)
        assert response.status_code == 503
        assert response.json["message"] == expected_no_auth_db_response

    # assert authenticated
    def test_attempt_fails_when_not_authenticated_or_token_expired(self, lowball_app_mock_providers,
                                                                   not_authenticated_or_token_expired_header):
        response = lowball_app_mock_providers.get(self.ROUTE, headers=not_authenticated_or_token_expired_header,
                                                  json={})
        assert response.status_code != 200
        assert any([
            "Token Has Expired" in response.json["message"],
            "No token provided" in response.json["message"],
            "Token Is Invalid" in response.json["message"]
        ])

    # success return full token
    def test_returns_decoded_token(self, lowball_app_mock_providers, token_request_rounds,
                                                     mock_lookup_token_filled):
        header, expected = token_request_rounds

        response = lowball_app_mock_providers.get(self.ROUTE, headers=header)
        assert response.status_code == 200
        assert response.json == expected.to_dict()


class TestGetCurrentClientTokens:
    ROUTE = "/builtins/auth/tokens"

    # test no auth db
    def test_fails_with_no_auth_database(self, lowball_app_no_db, expected_no_auth_db_response, basic_header):
        response = lowball_app_no_db.get(self.ROUTE, headers=basic_header)
        assert response.status_code == 503
        assert response.json["message"] == expected_no_auth_db_response

    # assert authenticated
    def test_attempt_fails_when_not_authenticated_or_token_expired(self, lowball_app_mock_providers,
                                                                   not_authenticated_or_token_expired_header):
        response = lowball_app_mock_providers.get(self.ROUTE, headers=not_authenticated_or_token_expired_header,
                                                  json={})
        assert response.status_code != 200
        assert any([
            "Token Has Expired" in response.json["message"],
            "No token provided" in response.json["message"],
            "Token Is Invalid" in response.json["message"]
        ])

    def test_list_tokens_gets_user_tokens_with_no_filters_applied(self,
                                                               lowball_app_mock_providers,
                                                               mock_lookup_token_filled,
                                                               normal1_token1_request_header,
                                                               client_id_normal1,
                                                               mock_list_tokens_filled,
                                                               mock_list_tokens_by_client_id,
                                                               non_admin_client_list_token_expected_response
                                                               ):
        response = lowball_app_mock_providers.get(f"{self.ROUTE}", headers=normal1_token1_request_header)
        assert isinstance(response.json, list)
        assert response.status_code == 200
        assert all(token_dict in response.json for token_dict in non_admin_client_list_token_expected_response)
        assert all(token_dict in non_admin_client_list_token_expected_response for token_dict in response.json)
        lowball_app_mock_providers.application.auth_db.list_tokens_by_client_id.assert_called_once_with(client_id_normal1)

    def test_list_tokens_gets_correct_tokens_allowed_filters_applied(self,
                                                                      lowball_app_mock_providers,
                                                                      mock_lookup_token_filled,
                                                                      normal1_token1_request_header,
                                                                      client_id_normal1,
                                                                      mock_list_tokens_filled,
                                                                      mock_list_tokens_by_client_id,
                                                                      non_admin_client_list_token_expected_response_with_queries

                                                                      ):
        query, expected_response = non_admin_client_list_token_expected_response_with_queries

        response = lowball_app_mock_providers.get(f"{self.ROUTE}{query}", headers=normal1_token1_request_header)
        assert isinstance(response.json, list)
        assert response.status_code == 200
        assert all(token_dict in response.json for token_dict in expected_response)
        assert all(token_dict in expected_response for token_dict in response.json)
        lowball_app_mock_providers.application.auth_db.list_tokens_by_client_id.assert_called_once_with(client_id_normal1)


    def test_list_tokens_fails_when_invalid_types_returned_from_auth_db(self, mock_list_tokens_returns_bad_values,
                                                                        lowball_app_mock_providers,
                                                                        mock_lookup_token_filled,
                                                                        normal1_token1_request_header,
                                                                        client_id_normal1
                                                                        ):
        normal_response = lowball_app_mock_providers.get(f"{self.ROUTE}", headers=normal1_token1_request_header)

        assert normal_response.status_code == 500
        assert "auth_db.list_tokens did not return a list as expected" in normal_response.json["message"] or \
               "auth_db.list_tokens returned a list that included a non-Token object" in normal_response.json["message"]
        lowball_app_mock_providers.application.auth_db.list_tokens_by_client_id.assert_called_once_with(client_id_normal1)


class TestCreateToken:
    ROUTE = "/builtins/auth/tokens"

    # assert authenticated
    def test_attempt_fails_when_not_authenticated_or_token_expired(self, lowball_app_mock_providers,
                                                                   not_authenticated_or_token_expired_header):
        response = lowball_app_mock_providers.post(self.ROUTE, headers=not_authenticated_or_token_expired_header,
                                                   json={})
        assert response.status_code != 201
        assert any([
            "Token Has Expired" in response.json["message"],
            "No token provided" in response.json["message"],
            "Token Is Invalid" in response.json["message"]
        ])

        # test no auth db

    def test_fails_with_no_auth_database(self, lowball_app_no_db, expected_no_auth_db_response, basic_header):
        response = lowball_app_no_db.post(self.ROUTE, headers=basic_header)
        assert response.status_code == 503
        assert response.json["message"] == expected_no_auth_db_response

    def test_failure_when_no_auth_provider_initialized_for_non_admin_token(self,
                                                                           lowball_app_no_auth_provider,
                                                                           expected_no_auth_provider_error_response,
                                                                           normal1_token1_request_header,
                                                                           mock_lookup_token_filled):
        response = lowball_app_no_auth_provider.post(self.ROUTE, headers=normal1_token1_request_header)
        print(response.json)
        assert response.status_code == 503
        assert response.json["message"] == expected_no_auth_provider_error_response

    def test_create_token_fails_for_invalid_token_life(self,
                                                       create_token_request_bad_token_life_rounds,
                                                       lowball_app_mock_providers,
                                                       mock_lookup_token_filled,
                                                       admin1_token1_request_header,
                                                       expected_bad_token_life_response
                                                       ):
        response = lowball_app_mock_providers.post(self.ROUTE, headers=admin1_token1_request_header,
                                                   json=create_token_request_bad_token_life_rounds)
        assert response.status_code == 400
        assert expected_bad_token_life_response in response.json["message"]

    def test_create_token_fails_for_non_admin_client_when_request_is_for_another_client(self,
                                                                                        normal1_token1_request_header,
                                                                                        client_id_normal2,
                                                                                        lowball_app_mock_providers,
                                                                                        mock_lookup_token_filled
                                                                                        ):
        response = lowball_app_mock_providers.post(self.ROUTE, headers=normal1_token1_request_header,
                                                   json={"client_id": client_id_normal2})

        assert response.status_code == 401
        assert "Current Token Has Inadequate Roles for Requested Action" in response.json["message"]

    def test_create_token_fails_for_non_admin_client_when_get_client_is_not_implemented(self,
                                                                                        normal1_token1_request_header,
                                                                                        client_id_normal1,
                                                                                        lowball_app_mock_providers,
                                                                                        mock_lookup_token_filled,
                                                                                        mock_get_client_not_implemented
                                                                                        ):
        response = lowball_app_mock_providers.post(self.ROUTE, headers=normal1_token1_request_header,
                                                   json={"client_id": client_id_normal1})
        lowball_app_mock_providers.application.auth_provider.get_client.assert_called_once_with(client_id_normal1)

        print(response.json)
        assert response.status_code == 409
        assert "get_client is not implemented in this auth provider. Non admin clients are unable to create tokens" \
               in response.json["message"]

    def test_create_token_creates_expected_tokens_for_admin_clients(self,
                                                                    lowball_app_mock_providers,
                                                                    mock_lookup_token_filled,
                                                                    admin1_token1_request_header,
                                                                    admin_client_create_token_success_rounds,
                                                                    mock_add_token
                                                                    ):
        request_body, expected_token_string, expected_token_object = admin_client_create_token_success_rounds
        print(expected_token_object.to_dict())
        response = lowball_app_mock_providers.post(self.ROUTE, headers=admin1_token1_request_header, json=request_body)

        assert response.status_code == 201
        assert response.json["token"] == expected_token_string
        assert response.json["token_data"] == expected_token_object.to_dict()
        lowball_app_mock_providers.application.authenticator.create_token.assert_called_once_with(
            expected_token_object.client_id,
            expected_token_object.roles,
            expected_token_object.issued_by,
            expected_token_object.expiration
        )
        lowball_app_mock_providers.application.auth_db.add_token.assert_called_once_with(expected_token_object)

    def test_create_token_creates_expected_tokens_for_non_admin_clients(self,
                                                                        lowball_app_mock_providers,
                                                                        mock_lookup_token_filled,
                                                                        normal1_token1_request_header,
                                                                        normal_client_create_token_success_rounds,
                                                                        mock_add_token,
                                                                        mock_get_client_filled
                                                                        ):
        request_body, expected_token_string, expected_token_object = normal_client_create_token_success_rounds
        print(expected_token_object.to_dict())
        response = lowball_app_mock_providers.post(self.ROUTE, headers=normal1_token1_request_header, json=request_body)

        assert response.status_code == 201
        assert response.json["token"] == expected_token_string
        assert response.json["token_data"] == expected_token_object.to_dict()
        lowball_app_mock_providers.application.authenticator.create_token.assert_called_once_with(
            expected_token_object.client_id,
            expected_token_object.roles,
            expected_token_object.issued_by,
            expected_token_object.expiration
        )
        lowball_app_mock_providers.application.auth_db.add_token.assert_called_once_with(expected_token_object)

    def test_create_token_fails_for_non_admin_client_for_themselves_if_client_is_not_in_auth_provider(self,
                                                                                                      lowball_app_mock_providers,
                                                                                                      mock_lookup_token_filled,
                                                                                                      normal2_token1_request_header,
                                                                                                      client_id_normal2,
                                                                                                      mock_get_client_filled
                                                                                                      ):
        response = lowball_app_mock_providers.post(self.ROUTE, headers=normal2_token1_request_header)
        lowball_app_mock_providers.application.auth_provider.get_client.assert_called_once_with(client_id_normal2)
        assert response.status_code == 404
        assert "Client ID for requesting token not found in auth provider" in response.json["message"]

    def test_create_token_fails_for_non_admin_client_if_roles_not_in_clients_listed_roles(self,
                                                                                          lowball_app_mock_providers,
                                                                                          mock_lookup_token_filled,
                                                                                          normal1_token1_request_header,
                                                                                          client_id_normal1,
                                                                                          mock_get_client_filled,
                                                                                          base_role2
                                                                                          ):
        response = lowball_app_mock_providers.post(self.ROUTE, headers=normal1_token1_request_header,
                                                   json={"roles": base_role2})
        lowball_app_mock_providers.application.auth_provider.get_client.assert_called_once_with(client_id_normal1)
        assert response.status_code == 401
        assert "Current Token Has Inadequate Roles for Requested Action" in response.json["message"]


class TestDeleteCurrentClientTokens:
    ROUTE = "/builtins/auth/tokens"

    def test_fails_with_no_auth_database(self, lowball_app_no_db, expected_no_auth_db_response, basic_header):
        response = lowball_app_no_db.delete(self.ROUTE, headers=basic_header)
        assert response.status_code == 503
        assert response.json["message"] == expected_no_auth_db_response

    # assert authenticated
    def test_attempt_fails_when_not_authenticated_or_token_expired(self, lowball_app_mock_providers,
                                                                   not_authenticated_or_token_expired_header):
        response = lowball_app_mock_providers.delete(self.ROUTE, headers=not_authenticated_or_token_expired_header,
                                                     json={})
        assert response.status_code != 200
        assert any([
            "Token Has Expired" in response.json["message"],
            "No token provided" in response.json["message"],
            "Token Is Invalid" in response.json["message"]
        ])

    def test_client_calls_revoke_token_for_each_token_they_have(self, lowball_app_mock_providers,
                                                                          normal1_token1_request_header,
                                                                          client_id_normal1,
                                                                          mock_lookup_token_filled,
                                                                          mock_list_tokens_by_client_id,
                                                                          mock_revoke_token,
                                                                          token1_normal1,
                                                                          token2_normal1,
                                                                          token3_normal1_expired):
        response = lowball_app_mock_providers.delete(self.ROUTE, headers=normal1_token1_request_header,
                                                     json={})
        assert response.status_code == 204
        calls = [
            call(token1_normal1.token_id),
            call(token2_normal1.token_id),
            call(token3_normal1_expired.token_id)
        ]
        lowball_app_mock_providers.application.auth_db.list_tokens_by_client_id.assert_called_once_with(client_id_normal1)
        lowball_app_mock_providers.application.auth_db.revoke_token.assert_has_calls(calls)

    def test_list_tokens_fails_when_invalid_types_returned_from_auth_db(self, mock_list_tokens_returns_bad_values,
                                                                        lowball_app_mock_providers,
                                                                        mock_lookup_token_filled,
                                                                        normal1_token1_request_header,
                                                                        client_id_normal1
                                                                        ):
        normal_response = lowball_app_mock_providers.delete(f"{self.ROUTE}", headers=normal1_token1_request_header)

        assert normal_response.status_code == 500
        assert "auth_db.list_tokens did not return a list as expected" in normal_response.json["message"] or \
               "auth_db.list_tokens returned a list that included a non-Token object" in normal_response.json["message"]
        lowball_app_mock_providers.application.auth_db.list_tokens_by_client_id.assert_called_once_with(client_id_normal1)


class TestGetAllTokens:
    ROUTE = "/builtins/auth/tokens/all"

    def test_fails_with_no_auth_database(self, lowball_app_no_db, expected_no_auth_db_response, basic_header):
        response = lowball_app_no_db.get(self.ROUTE, headers=basic_header)
        assert response.status_code == 503
        assert response.json["message"] == expected_no_auth_db_response

    def test_attempt_fails_when_not_authenticated_or_token_expired(self, lowball_app_mock_providers,
                                                                               not_authenticated_or_token_expired_header):
        response = lowball_app_mock_providers.get(self.ROUTE, headers=not_authenticated_or_token_expired_header,
                                                    json={})
        assert response.status_code != 200
        assert any([
            "Token Has Expired" in response.json["message"],
            "No token provided" in response.json["message"],
            "Token Is Invalid" in response.json["message"]
        ])

    # assert authenticated admin
    def test_attempt_fails_when_authenticated_but_not_admin(self, lowball_app_mock_providers,
                                                                               non_admin_headers):
        response = lowball_app_mock_providers.get(self.ROUTE, headers=non_admin_headers,
                                                    json={})
        assert response.status_code == 401
        assert "Unauthorized" in response.json["message"]

    def test_list_tokens_for_admin_client_gets_all_tokens_when_no_filters_applied(self, lowball_app_mock_providers,
                                                                                  mock_lookup_token_filled,
                                                                                  token_db, admin1_token1_request_header,
                                                                                  mock_list_tokens_filled):

        response = lowball_app_mock_providers.get(self.ROUTE, headers=admin1_token1_request_header)

        assert response.status_code == 200
        print(response.json)
        assert all(token.to_dict() in response.json for token in token_db.values())
        assert all(token_dict in [token.to_dict() for token in token_db.values()] for token_dict in response.json)
        lowball_app_mock_providers.application.auth_db.list_tokens.assert_called_once()

    def test_list_tokens_for_admin_client_gets_correct_tokens_when_role_filters_applied(self, base_role1,
                                                                                        lowball_app_mock_providers,
                                                                                        mock_lookup_token_filled,
                                                                                        admin_list_token_roles_request_response,
                                                                                        admin1_token1_request_header,
                                                                                        mock_list_tokens_filled):

        query, expected_response = admin_list_token_roles_request_response

        response = lowball_app_mock_providers.get(f"{self.ROUTE}{query}", headers=admin1_token1_request_header)
        assert isinstance(response.json, list)
        assert response.status_code == 200
        assert all(token_dict in response.json for token_dict in expected_response)
        assert all(token_dict in expected_response for token_dict in response.json)
        lowball_app_mock_providers.application.auth_db.list_tokens.assert_called_once()

    def test_list_tokens_for_admin_client_gets_correct_tokens_when_client_filters_applied(self,
                                                                                          lowball_app_mock_providers,
                                                                                          mock_lookup_token_filled,
                                                                                          admin_list_token_client_request_response,
                                                                                          admin1_token1_request_header,
                                                                                          mock_list_tokens_filled
                                                                                          ):
        query, expected_response = admin_list_token_client_request_response

        response = lowball_app_mock_providers.get(f"{self.ROUTE}{query}", headers=admin1_token1_request_header)
        assert isinstance(response.json, list)
        assert response.status_code == 200
        assert all(token_dict in response.json for token_dict in expected_response)
        assert all(token_dict in expected_response for token_dict in response.json)
        lowball_app_mock_providers.application.auth_db.list_tokens.assert_called_once()

    def test_list_tokens_for_admin_client_gets_correct_tokens_when_exclude_expired_filter_set_to_true(self,
                                                                                                      lowball_app_mock_providers,
                                                                                                      mock_lookup_token_filled,
                                                                                                      admin1_token1_request_header,
                                                                                                      mock_list_tokens_filled,
                                                                                                      token_db_admin_exclude_expired_response
                                                                                                      ):

        query = "?exclude_expired=yes"
        response = lowball_app_mock_providers.get(f"{self.ROUTE}{query}", headers=admin1_token1_request_header)
        assert isinstance(response.json, list)
        assert response.status_code == 200
        assert all(token_dict in response.json for token_dict in token_db_admin_exclude_expired_response)
        assert all(token_dict in token_db_admin_exclude_expired_response for token_dict in response.json)
        lowball_app_mock_providers.application.auth_db.list_tokens.assert_called_once()

    def test_list_tokens_for_admin_client_gets_correct_tokens_when_multiple_filters_applied(self,
                                                                                            lowball_app_mock_providers,
                                                                                            mock_lookup_token_filled,
                                                                                            admin1_token1_request_header,
                                                                                            mock_list_tokens_filled,
                                                                                            admin_list_token_multi_query_request_response
                                                                                            ):
        query, expected_response = admin_list_token_multi_query_request_response

        response = lowball_app_mock_providers.get(f"{self.ROUTE}{query}", headers=admin1_token1_request_header)
        assert isinstance(response.json, list)
        assert response.status_code == 200
        assert all(token_dict in response.json for token_dict in expected_response)
        assert all(token_dict in expected_response for token_dict in response.json)
        lowball_app_mock_providers.application.auth_db.list_tokens.assert_called_once()

    def test_list_tokens_fails_when_invalid_types_returned_from_auth_db(self, mock_list_tokens_returns_bad_values,
                                                                        lowball_app_mock_providers,
                                                                        mock_lookup_token_filled,
                                                                        admin1_token1_request_header
                                                                        ):
        normal_response = lowball_app_mock_providers.get(f"{self.ROUTE}", headers=admin1_token1_request_header)

        assert normal_response.status_code == 500
        assert "auth_db.list_tokens did not return a list as expected" in normal_response.json["message"] or \
               "auth_db.list_tokens returned a list that included a non-Token object" in normal_response.json["message"]
        lowball_app_mock_providers.application.auth_db.list_tokens.assert_called_once()


class TestDeleteAllTokens:
    ROUTE = "/builtins/auth/tokens/all"

    def test_fails_with_no_auth_database(self, lowball_app_no_db, expected_no_auth_db_response, basic_header):
        response = lowball_app_no_db.delete(self.ROUTE, headers=basic_header)
        assert response.status_code == 503
        assert response.json["message"] == expected_no_auth_db_response

    def test_attempt_fails_when_not_authenticated_or_token_expired(self, lowball_app_mock_providers,
                                                                   not_authenticated_or_token_expired_header):
        response = lowball_app_mock_providers.delete(self.ROUTE, headers=not_authenticated_or_token_expired_header,
                                                  json={})
        assert response.status_code != 200
        assert any([
            "Token Has Expired" in response.json["message"],
            "No token provided" in response.json["message"],
            "Token Is Invalid" in response.json["message"]
        ])

    # assert authenticated admin
    def test_attempt_fails_when_authenticated_but_not_admin(self, lowball_app_mock_providers,
                                                            non_admin_headers):
        response = lowball_app_mock_providers.delete(self.ROUTE, headers=non_admin_headers,
                                                  json={})
        assert response.status_code == 401
        assert "Unauthorized" in response.json["message"]

    def test_calls_revoke_all_when_no_filters_supplied(self, lowball_app_mock_providers,
                                                       admin1_token1_request_header,
                                                       mock_lookup_token_filled,
                                                       mock_list_tokens_filled,
                                                       mock_list_tokens_by_client_id,
                                                       mock_revoke_all
                                                       ):
        response = lowball_app_mock_providers.delete(self.ROUTE, headers=admin1_token1_request_header,
                                                     json={})
        assert response.status_code == 204
        lowball_app_mock_providers.application.auth_db.revoke_all.assert_called_once()

    def test_calls_revoke_token_with_correct_tokens_client_filter_applied(self,
                                                                          lowball_app_mock_providers,
                                                                          admin1_token1_request_header,
                                                                          mock_lookup_token_filled,
                                                                          mock_list_tokens_filled,
                                                                          admin_delete_token_client_request_response,
                                                                          mock_revoke_token
                                                                          ):
        query, expected_tokens = admin_delete_token_client_request_response
        response = lowball_app_mock_providers.delete(f"{self.ROUTE}{query}", headers=admin1_token1_request_header,
                                                     json={})
        assert response.status_code == 204
        calls = [
            call(token.token_id) for token in expected_tokens
        ]
        lowball_app_mock_providers.application.auth_db.list_tokens.assert_called_once()
        lowball_app_mock_providers.application.auth_db.revoke_token.assert_has_calls(calls)

    def test_calls_revoke_token_with_correct_tokens_role_filter_applied(self,
                                                                          lowball_app_mock_providers,
                                                                          admin1_token1_request_header,
                                                                          mock_lookup_token_filled,
                                                                          mock_list_tokens_filled,
                                                                          admin_delete_token_roles_request_response,
                                                                          mock_revoke_token
                                                                          ):
        query, expected_tokens = admin_delete_token_roles_request_response
        response = lowball_app_mock_providers.delete(f"{self.ROUTE}{query}", headers=admin1_token1_request_header,
                                                     json={})
        assert response.status_code == 204
        calls = [
            call(token.token_id) for token in expected_tokens
        ]
        lowball_app_mock_providers.application.auth_db.list_tokens.assert_called_once()
        lowball_app_mock_providers.application.auth_db.revoke_token.assert_has_calls(calls)

    def test_calls_revoke_token_with_correct_tokens_multiple_filters(self, lowball_app_mock_providers,
                                                                          admin1_token1_request_header,
                                                                          mock_lookup_token_filled,
                                                                          mock_list_tokens_filled,
                                                                          admin_delete_token_multi_query_request_response,
                                                                          mock_revoke_token
                                                                          ):
        query, expected_tokens = admin_delete_token_multi_query_request_response
        response = lowball_app_mock_providers.delete(f"{self.ROUTE}{query}", headers=admin1_token1_request_header)

        assert response.status_code == 204
        calls = [
            call(token.token_id) for token in expected_tokens
        ]
        lowball_app_mock_providers.application.auth_db.list_tokens.assert_called_once()
        lowball_app_mock_providers.application.auth_db.revoke_token.assert_has_calls(calls)

    def test_list_tokens_fails_when_invalid_types_returned_from_auth_db(self, mock_list_tokens_returns_bad_values,
                                                                        lowball_app_mock_providers,
                                                                        mock_lookup_token_filled,
                                                                        admin1_token1_request_header,
                                                                        admin_delete_token_multi_query_request_response
                                                                        ):
        query, expected_tokens = admin_delete_token_multi_query_request_response
        normal_response = lowball_app_mock_providers.delete(f"{self.ROUTE}{query}", headers=admin1_token1_request_header)

        assert normal_response.status_code == 500
        assert "auth_db.list_tokens did not return a list as expected" in normal_response.json["message"] or \
               "auth_db.list_tokens returned a list that included a non-Token object" in normal_response.json["message"]
        lowball_app_mock_providers.application.auth_db.list_tokens.assert_called_once()


class TestCleanupTokens:
    ROUTE = "/builtins/auth/tokens/cleanup"

    def test_fails_with_no_auth_database(self, lowball_app_no_db, expected_no_auth_db_response, basic_header):
        response = lowball_app_no_db.post(self.ROUTE, headers=basic_header)
        assert response.status_code == 503
        assert response.json["message"] == expected_no_auth_db_response


    # assert authenticated
    def test_attempt_fails_when_not_authenticated_or_token_expired(self, lowball_app_mock_providers,
                                                                   not_authenticated_or_token_expired_header):
        response = lowball_app_mock_providers.post(self.ROUTE, headers=not_authenticated_or_token_expired_header,
                                                   json={})
        assert response.status_code != 200
        assert any([
            "Token Has Expired" in response.json["message"],
            "No token provided" in response.json["message"],
            "Token Is Invalid" in response.json["message"]
        ])

    # assert authenticated admin
    def test_attempt_fails_when_authenticated_but_not_admin(self, lowball_app_mock_providers,
                                                            non_admin_headers):
        response = lowball_app_mock_providers.post(self.ROUTE, headers=non_admin_headers,
                                                   json={})
        assert response.status_code == 401
        assert "Unauthorized" in response.json["message"]

    # success route
    def test_cleanup_token_calls_the_cleanup_tokens_function_on_success(self,
                                                                        lowball_app_mock_providers,
                                                                        admin1_token1_request_header,
                                                                        mock_lookup_token_filled,
                                                                        mock_cleanup_tokens
                                                                        ):
        response = lowball_app_mock_providers.post(self.ROUTE, headers=admin1_token1_request_header)
        assert response.status_code == 204
        lowball_app_mock_providers.application.auth_db.cleanup_tokens.assert_called_once()


class TestGetToken:
    ROUTE = "/builtins/auth/tokens/{}"

    def test_fails_with_no_auth_database(self, lowball_app_no_db, expected_no_auth_db_response, basic_header):
        route = self.ROUTE.format("anytoken")
        response = lowball_app_no_db.get(route, headers=basic_header)
        assert response.status_code == 503
        assert response.json["message"] == expected_no_auth_db_response

    # assert authenticated
    def test_attempt_fails_when_not_authenticated_or_token_expired(self, lowball_app_mock_providers,
                                                                   token_id,
                                                                   not_authenticated_or_token_expired_header):
        route = self.ROUTE.format(token_id)
        response = lowball_app_mock_providers.get(route, headers=not_authenticated_or_token_expired_header)
        assert response.status_code != 200
        assert any([
            "Token Has Expired" in response.json["message"],
            "No token provided" in response.json["message"],
            "Token Is Invalid" in response.json["message"]
        ])

    def test_attempt_fails_when_invalid_token_id_format(self,
                                                        lowball_app_mock_providers,
                                                        mock_lookup_token_filled,
                                                        mock_list_tokens_filled,
                                                        invalid_token_ids,
                                                        admin1_token1_request_header
                                                        ):
        route = self.ROUTE.format(invalid_token_ids)
        response = lowball_app_mock_providers.get(route, headers=admin1_token1_request_header)

        assert response.status_code == 400
        assert "Invalid token id format" in response.json["message"]

    def test_not_found_when_token_id_not_found(self,
                                               lowball_app_mock_providers,
                                               mock_lookup_token_filled,
                                               mock_list_tokens_filled,
                                               admin1_token1_request_header
                                               ):
        route = self.ROUTE.format("amissingtokenidI")
        response = lowball_app_mock_providers.get(route, headers=admin1_token1_request_header)
        assert call("amissingtokenidI") in lowball_app_mock_providers.application.auth_db.lookup_token.mock_calls
        assert response.status_code == 404

    def test_admin_client_success_when_looking_up_token_for_other_client(self,
                                                                         lowball_app_mock_providers,
                                                                         mock_lookup_token_filled,
                                                                         admin1_token1_request_header,
                                                                         mock_list_tokens_filled,
                                                                         expected_lookup_token_response_for_admin
                                                                         ):
        token_id, expected_response = expected_lookup_token_response_for_admin
        route = self.ROUTE.format(token_id)
        response = lowball_app_mock_providers.get(route, headers=admin1_token1_request_header)
        assert call(token_id) in lowball_app_mock_providers.application.auth_db.lookup_token.mock_calls
        assert response.status_code == 200
        assert response.json == expected_response

    def test_non_admin_client_failure_when_looking_up_token_for_other_client(self,
                                                                             lowball_app_mock_providers,
                                                                             mock_lookup_token_filled,
                                                                             mock_list_tokens_filled,
                                                                             normal1_token1_request_header,
                                                                             token2_normal2
                                                                             ):
        route = self.ROUTE.format(token2_normal2.token_id)
        response = lowball_app_mock_providers.get(route, headers=normal1_token1_request_header)
        assert response.status_code == 401
        assert "Current Token Has Inadequate Roles for Requested Action" in response.json["message"]

    def test_success_when_client_looks_up_token_owned_by_client(self,
                                                                lowball_app_mock_providers,
                                                                mock_lookup_token_filled,
                                                                mock_list_tokens_filled,
                                                                normal1_token1_request_header,
                                                                normal1_token_ids_request_response
                                                                ):
        token_id, expected_response = normal1_token_ids_request_response
        route = self.ROUTE.format(token_id)
        response = lowball_app_mock_providers.get(route, headers=normal1_token1_request_header)

        assert call(token_id) in lowball_app_mock_providers.application.auth_db.lookup_token.mock_calls
        assert response.status_code == 200
        assert response.json == expected_response


class TestDeleteToken:
    ROUTE = "/builtins/auth/tokens/{}"

    def test_fails_with_no_auth_database(self, lowball_app_no_db, expected_no_auth_db_response, basic_header):
        response = lowball_app_no_db.delete(self.ROUTE.format("anytokenid"), headers=basic_header)
        assert response.status_code == 503
        assert response.json["message"] == expected_no_auth_db_response

    # assert authenticated
    def test_attempt_fails_when_not_authenticated_or_token_expired(self, lowball_app_mock_providers,
                                                                   not_authenticated_or_token_expired_header,
                                                                   token_id):
        route = self.ROUTE.format(token_id)
        response = lowball_app_mock_providers.delete(route, headers=not_authenticated_or_token_expired_header,
                                                     json={})
        assert response.status_code != 200
        assert any([
            "Token Has Expired" in response.json["message"],
            "No token provided" in response.json["message"],
            "Token Is Invalid" in response.json["message"]
        ])

    def test_attempt_fails_when_invalid_token_id_format(self,
                                                        lowball_app_mock_providers,
                                                        mock_lookup_token_filled,
                                                        mock_list_tokens_filled,
                                                        invalid_token_ids,
                                                        admin1_token1_request_header
                                                        ):
        route = self.ROUTE.format(invalid_token_ids)
        response = lowball_app_mock_providers.delete(route, headers=admin1_token1_request_header)

        assert response.status_code == 400
        assert "Invalid token id format" in response.json["message"]

    # admin
    def test_not_found_when_token_id_is_not_found(self,
                                                  lowball_app_mock_providers,
                                                  mock_lookup_token_filled,
                                                  mock_list_tokens_filled,
                                                  admin1_token1_request_header
                                                  ):
        route = self.ROUTE.format("amissingtokenidI")
        response = lowball_app_mock_providers.delete(route, headers=admin1_token1_request_header)

        assert response.status_code == 404

    def test_attempt_succeeds_for_admin_client_with_any_token(self,
                                                              lowball_app_mock_providers,
                                                              mock_lookup_token_filled,
                                                              admin1_token1_request_header,
                                                              mock_list_tokens_filled,
                                                              expected_lookup_token_response_for_admin,
                                                              mock_revoke_token
                                                              ):
        token_id, expected_response = expected_lookup_token_response_for_admin
        route = self.ROUTE.format(token_id)
        response = lowball_app_mock_providers.delete(route, headers=admin1_token1_request_header)
        lowball_app_mock_providers.application.auth_db.revoke_token.assert_called_once_with(token_id)
        assert response.status_code == 204

    def test_attempt_unauthorized_for_non_admin_client_with_non_owned_token(self,
                                                                            lowball_app_mock_providers,
                                                                            mock_lookup_token_filled,
                                                                            mock_list_tokens_filled,
                                                                            normal1_token1_request_header,
                                                                            token2_normal2
                                                                            ):
        route = self.ROUTE.format(token2_normal2.token_id)
        response = lowball_app_mock_providers.delete(route, headers=normal1_token1_request_header)
        assert response.status_code == 401
        assert "Current Token Has Inadequate Roles for Requested Action" in response.json["message"]

    def test_attempt_success_for_any_client_with_owned_token(self,
                                                             lowball_app_mock_providers,
                                                             mock_lookup_token_filled,
                                                             mock_list_tokens_filled,
                                                             normal1_token1_request_header,
                                                             normal1_token_ids_request_response,
                                                             mock_revoke_token
                                                             ):
        token_id, expected_response = normal1_token_ids_request_response
        route = self.ROUTE.format(token_id)
        response = lowball_app_mock_providers.delete(route, headers=normal1_token1_request_header)
        lowball_app_mock_providers.application.auth_db.revoke_token.assert_called_once_with(token_id)

        assert response.status_code == 204


class TestGetCurrentClient:
    ROUTE = "/builtins/auth/clients"

    def test_fails_with_no_auth_database(self, lowball_app_no_db, expected_no_auth_db_response, basic_header):
        response = lowball_app_no_db.get(self.ROUTE, headers=basic_header)
        assert response.status_code == 503
        assert response.json["message"] == expected_no_auth_db_response

    # assert authenticated
    def test_attempt_fails_when_not_authenticated_or_token_expired(self, lowball_app_mock_providers,
                                                                   not_authenticated_or_token_expired_header,
                                                                   client_id_normal1):
        response = lowball_app_mock_providers.get(self.ROUTE, headers=not_authenticated_or_token_expired_header,
                                                  json={})
        assert response.status_code != 200
        assert any([
            "Token Has Expired" in response.json["message"],
            "No token provided" in response.json["message"],
            "Token Is Invalid" in response.json["message"]
        ])

    def test_failure_when_no_auth_provider_initialized(self,
                                                       lowball_app_no_auth_provider,
                                                       expected_no_auth_provider_error_response,
                                                       admin1_token1_request_header,
                                                       mock_lookup_token_filled):
        response = lowball_app_no_auth_provider.get(self.ROUTE, headers=admin1_token1_request_header)
        assert response.status_code == 503
        assert response.json["message"] == expected_no_auth_provider_error_response

    def test_not_found_if_client_not_found(self,
                                           lowball_app_mock_providers,
                                           normal2_token1_request_header,
                                           client_id_normal2,
                                           mock_lookup_token_filled,
                                           mock_get_client_filled
                                           ):
        response = lowball_app_mock_providers.get(self.ROUTE, headers=normal2_token1_request_header, json={})
        print(response.json)
        assert response.status_code == 404
        lowball_app_mock_providers.application.auth_provider.get_client.assert_called_once_with(client_id_normal2)

    def test_success_outputs_for_client_data(self,
                                             lowball_app_mock_providers,
                                             mock_lookup_token_filled,
                                             mock_get_client_filled,
                                             get_current_client_request_response_rounds
                                             ):
        header, client_id,expected_response = get_current_client_request_response_rounds

        response = lowball_app_mock_providers.get(self.ROUTE, headers=header)
        print(response.json)
        assert response.status_code == 200
        assert response.json == expected_response
        lowball_app_mock_providers.application.auth_provider.get_client.assert_called_once_with(client_id)


class TestClientSelfUpdate:
    ROUTE = "/builtins/auth/clients"

    def test_fails_with_no_auth_database(self, lowball_app_no_db, expected_no_auth_db_response, basic_header):
        response = lowball_app_no_db.post(self.ROUTE, headers=basic_header)
        assert response.status_code == 503
        assert response.json["message"] == expected_no_auth_db_response

    # assert authenticated
    def test_self_update_attempt_fails_when_not_authenticated_or_token_expired(self, lowball_app_mock_providers,
                                                                               not_authenticated_or_token_expired_header):
        response = lowball_app_mock_providers.post(self.ROUTE, headers=not_authenticated_or_token_expired_header,
                                                    json={})
        assert response.status_code != 200
        assert any([
            "Token Has Expired" in response.json["message"],
            "No token provided" in response.json["message"],
            "Token Is Invalid" in response.json["message"]
        ])

    # fail client not found
    def test_self_update_attempt_fails_when_client_not_found(self,
                                                             lowball_app_mock_providers,
                                                             basic_header,
                                                             mock_get_client_not_found,
                                                             mock_decode_token,
                                                             mock_lookup_token,
                                                             mock_self_update_class,
                                                             ):
        response = lowball_app_mock_providers.post(self.ROUTE, headers=basic_header)
        assert response.status_code == 404
        assert "Client not found" in response.json["message"]

    # test no auth provider
    def test_self_update_attempt_fails_with_no_auth_provider_present(
            self, lowball_app_no_auth_provider, expected_no_auth_provider_error_response, basic_header,
            mock_decode_token, mock_lookup_token):
        response = lowball_app_no_auth_provider.post(self.ROUTE, headers=basic_header)
        print(response.json)
        assert response.status_code == 503
        assert response.json["message"] == expected_no_auth_provider_error_response

    # test no client self update package class
    def test_self_update_attempt_fails_with_no_self_update_package_class_defined(self,
                                                                                 lowball_app_mock_providers,
                                                                                 basic_header,
                                                                                 mock_decode_token,
                                                                                 mock_lookup_token,
                                                                                 mock_get_client_found
                                                                                 ):
        response = lowball_app_mock_providers.post(self.ROUTE, headers=basic_header)

        assert response.status_code == 501
        assert response.json["message"] == "501 Not Implemented: self_update_client_package_class not implemented"

    # test bad json/patch data
    def test_self_update_attempt_fails_with_invalid_patch_data(self,
                                                               lowball_app_mock_providers,
                                                               basic_header,
                                                               mock_decode_token,
                                                               mock_lookup_token,
                                                               mock_self_update_class,
                                                               expected_bad_package_data_error_response,
                                                               does_not_match_self_update_package_request,
                                                               mock_get_client_found):
        response = lowball_app_mock_providers.post(self.ROUTE,
                                                    headers=basic_header,
                                                    json=does_not_match_self_update_package_request)

        assert response.status_code == 400
        assert response.json["message"] == expected_bad_package_data_error_response

    def test_self_update_success_flow(self,
                                      lowball_app_mock_providers,
                                      basic_header,
                                      mock_decode_token,
                                      mock_lookup_token,
                                      mock_self_update_class,
                                      basic_client_data,
                                      mock_self_update_rounds,
                                      success_request,
                                      mock_get_client_found
                                      ):
        response = lowball_app_mock_providers.post(self.ROUTE, json=success_request, headers=basic_header)
        print(response.json)
        assert response.status_code == 200
        assert response.json == basic_client_data.to_dict()
        lowball_app_mock_providers.application.auth_provider.client_self_update.assert_called_once()


class TestCreateClient:
    ROUTE = "/builtins/auth/clients/create"

    def test_fails_with_no_auth_database(self, lowball_app_no_db, expected_no_auth_db_response, basic_header):
        response = lowball_app_no_db.post(self.ROUTE, headers=basic_header)
        assert response.status_code == 503
        assert response.json["message"] == expected_no_auth_db_response

    # assert authenticated
    def test_attempt_fails_when_not_authenticated_or_token_expired(self, lowball_app_mock_providers,
                                                                   not_authenticated_or_token_expired_header):
        response = lowball_app_mock_providers.post(self.ROUTE, headers=not_authenticated_or_token_expired_header,
                                                   json={})
        assert response.status_code != 200
        assert any([
            "Token Has Expired" in response.json["message"],
            "No token provided" in response.json["message"],
            "Token Is Invalid" in response.json["message"]
        ])

    # assert authenticated admin
    def test_attempt_fails_when_authenticated_but_not_admin(self, lowball_app_mock_providers,
                                                            non_admin_headers):
        response = lowball_app_mock_providers.post(self.ROUTE, headers=non_admin_headers,
                                                   json={})
        assert response.status_code == 401
        assert "Unauthorized" in response.json["message"]

    # fail no auth provider
    def test_failure_when_no_auth_provider_initialized(self,
                                                       lowball_app_no_auth_provider,
                                                       expected_no_auth_provider_error_response,
                                                       admin1_token1_request_header,
                                                       mock_lookup_token_filled):
        response = lowball_app_no_auth_provider.post(self.ROUTE, headers=admin1_token1_request_header)
        assert response.status_code == 503
        assert response.json["message"] == expected_no_auth_provider_error_response

    # fail no package class
    def test_attempt_fails_when_auth_provider_does_not_have_an_implemented_client_creation_package(
            self,
            lowball_app_mock_providers,
            admin1_token1_request_header,
            mock_lookup_token_filled,
    ):
        response = lowball_app_mock_providers.post(self.ROUTE, headers=admin1_token1_request_header, json={})
        print(response.json)
        assert response.status_code == 501
        assert "create_client_package_class" in response.json["message"]

    # fail bad input
    def test_attempt_fails_when_auth_provided_package_data_is_invalid(self,
                                                                      lowball_app_mock_providers,
                                                                      admin1_token1_request_header,
                                                                      mock_lookup_token_filled,
                                                                      mock_get_client_filled,
                                                                      expected_bad_package_data_error_response,
                                                                      does_not_match_update_package_request,
                                                                      mock_create_client_class
                                                                      ):
        response = lowball_app_mock_providers.post(self.ROUTE,
                                                   headers=admin1_token1_request_header,
                                                   json=does_not_match_update_package_request)

        assert response.status_code == 400
        assert response.json["message"] == expected_bad_package_data_error_response

    # success output
    def test_success_returns_expected_output(self,
                                             lowball_app_mock_providers,
                                             admin1_token1_request_header,
                                             mock_lookup_token_filled,
                                             basic_client_data,
                                             mock_create_client_rounds,
                                             success_request,
                                             mock_create_client_class
                                             ):
        response = lowball_app_mock_providers.post(self.ROUTE,
                                                   json=success_request,
                                                   headers=admin1_token1_request_header)
        print(response.json)
        assert response.status_code == 201
        assert response.json == basic_client_data.to_dict()
        lowball_app_mock_providers.application.auth_provider.create_client.assert_called_once()


class TestClientRegistration:
    ROUTE = "/builtins/auth/clients/register"

    # test no auth provider
    def test_self_registration_attempt_fails_with_no_auth_provider(
            self, lowball_app_no_auth_provider, expected_no_auth_provider_error_response):
        response = lowball_app_no_auth_provider.post(self.ROUTE)
        assert response.status_code == 503
        assert response.json["message"] == expected_no_auth_provider_error_response

    # test no package class
    def test_self_registration_attempt_fails_with_no_defined_registration_package_class(self,
                                                                                        lowball_app_mock_providers):
        response = lowball_app_mock_providers.post(self.ROUTE, json={"some": "json"})

        assert response.status_code == 501
        assert response.json["message"] == "501 Not Implemented: client_registration_package_class not implemented"

    # test not implemented
    def test_self_registration_attempt_fails_with_self_registration_not_implemented(self,
                                                                                    lowball_app_mock_providers,
                                                                                    mock_self_register_class,
                                                                                    success_request):
        response = lowball_app_mock_providers.post(self.ROUTE, json=success_request)
        assert response.status_code == 501
        assert response.json["message"] == "501 Not Implemented: client_self_register not implemented"

    # test bad json data
    def test_self_registration_attempt_fails_with_invalid_json_data(self,
                                                                    lowball_app_mock_providers,
                                                                    mock_self_register_class,
                                                                    does_not_match_self_register_package_request,
                                                                    expected_bad_package_data_error_response):
        response = lowball_app_mock_providers.post(self.ROUTE, json=does_not_match_self_register_package_request)
        assert response.status_code == 400
        assert response.json["message"] == expected_bad_package_data_error_response

    def test_self_registration_success_flow(self, lowball_app_mock_providers, mock_self_register_class,
                                            success_request, mock_self_register_rounds, basic_client_data):
        response = lowball_app_mock_providers.post(self.ROUTE, json=success_request)
        print(response.json)
        assert response.status_code == 201
        assert response.json == basic_client_data.to_dict()
        lowball_app_mock_providers.application.auth_provider.client_self_register.assert_called_once()


class TestGetAllClients:

    ROUTE = "/builtins/auth/clients/all"

    def test_fails_with_no_auth_database(self, lowball_app_no_db, expected_no_auth_db_response, basic_header):
        response = lowball_app_no_db.get(self.ROUTE, headers=basic_header)
        assert response.status_code == 503
        assert response.json["message"] == expected_no_auth_db_response

    # assert authenticated
    def test_attempt_fails_when_not_authenticated_or_token_expired(self, lowball_app_mock_providers,
                                                                   not_authenticated_or_token_expired_header):
        response = lowball_app_mock_providers.get(self.ROUTE, headers=not_authenticated_or_token_expired_header,
                                                   json={})
        assert response.status_code != 200
        assert any([
            "Token Has Expired" in response.json["message"],
            "No token provided" in response.json["message"],
            "Token Is Invalid" in response.json["message"]
        ])

    # assert authenticated admin
    def test_attempt_fails_when_authenticated_but_not_admin(self, lowball_app_mock_providers,
                                                            non_admin_headers):
        response = lowball_app_mock_providers.get(self.ROUTE, headers=non_admin_headers,
                                                   json={})
        assert response.status_code == 401
        assert "Unauthorized" in response.json["message"]

    def test_gets_correct_client_with_role_filter_applied(self,
                                                          lowball_app_mock_providers,
                                                          admin1_token1_request_header,
                                                          mock_lookup_token_filled,
                                                          mock_list_tokens_filled,
                                                          mock_list_clients_filled,
                                                          get_clients_request_response_rounds,
                                                          mock_revoke_token
                                                          ):
        query, expected_clients = get_clients_request_response_rounds
        response = lowball_app_mock_providers.get(f"{self.ROUTE}{query}", headers=admin1_token1_request_header,
                                                     json={})
        assert response.status_code == 200
        assert isinstance(response.json, list)
        assert all(client.to_dict() in response.json for client in expected_clients)
        assert all(client_dict in [client.to_dict() for client in expected_clients] for client_dict in response.json)
        lowball_app_mock_providers.application.auth_provider.list_clients.assert_called_once()

    def test_gets_all_clients_when_no_filters_applied(self, lowball_app_mock_providers,
                                                      mock_lookup_token_filled,
                                                      token_db, admin1_token1_request_header,
                                                      mock_list_tokens_filled,
                                                      mock_list_clients_filled,
                                                      client_db):

        response = lowball_app_mock_providers.get(self.ROUTE, headers=admin1_token1_request_header)

        assert response.status_code == 200
        print(response.json)
        assert isinstance(response.json, list)
        assert all(client.to_dict() in response.json for client in client_db.values())
        assert all(client_dict in [client.to_dict() for client in client_db.values()] for client_dict in response.json)
        lowball_app_mock_providers.application.auth_provider.list_clients.assert_called_once()

    def test_get_clients_fails_when_invalid_types_returned_from_auth_provider(self, mock_list_clients_returns_bad_values,
                                                                        lowball_app_mock_providers,
                                                                        mock_lookup_token_filled,
                                                                        admin1_token1_request_header
                                                                        ):
        normal_response = lowball_app_mock_providers.get(self.ROUTE, headers=admin1_token1_request_header)

        assert normal_response.status_code == 500
        assert "auth_provider.list_clients did not return a list of client data objects as expected" in normal_response.json["message"]
        lowball_app_mock_providers.application.auth_provider.list_clients.assert_called_once()


class TestGetClient:

    ROUTE = "/builtins/auth/clients/{}"

    def test_fails_with_no_auth_database(self, lowball_app_no_db, expected_no_auth_db_response, basic_header):
        route = self.ROUTE.format("anyclientid")
        response = lowball_app_no_db.get(route, headers=basic_header)
        assert response.status_code == 503
        assert response.json["message"] == expected_no_auth_db_response

    # assert authenticated
    def test_attempt_fails_when_not_authenticated_or_token_expired(self, lowball_app_mock_providers,
                                                                   not_authenticated_or_token_expired_header,
                                                                   client_id_normal1):
        route = self.ROUTE.format(client_id_normal1)
        response = lowball_app_mock_providers.get(route, headers=not_authenticated_or_token_expired_header,
                                                  json={})
        assert response.status_code != 200
        assert any([
            "Token Has Expired" in response.json["message"],
            "No token provided" in response.json["message"],
            "Token Is Invalid" in response.json["message"]
        ])

    def test_attempt_fails_when_authenticated_but_not_admin(self, lowball_app_mock_providers,
                                                            non_admin_headers):
        response = lowball_app_mock_providers.get(self.ROUTE, headers=non_admin_headers,
                                                   json={})
        assert response.status_code == 401
        assert "Unauthorized" in response.json["message"]

    def test_failure_when_no_auth_provider_initialized(self,
                                                       lowball_app_no_auth_provider,
                                                       expected_no_auth_provider_error_response,
                                                       admin1_token1_request_header,
                                                       mock_lookup_token_filled):
        response = lowball_app_no_auth_provider.get(self.ROUTE, headers=admin1_token1_request_header)
        assert response.status_code == 503
        assert response.json["message"] == expected_no_auth_provider_error_response

    def test_not_found_if_client_not_found(self,
                                           lowball_app_mock_providers,
                                           admin1_token1_request_header,
                                           client_id_normal2,
                                           mock_lookup_token_filled,
                                           mock_get_client_filled
                                           ):
        route = self.ROUTE.format(client_id_normal2)
        response = lowball_app_mock_providers.get(route, headers=admin1_token1_request_header, json={})
        print(response.json)
        assert response.status_code == 404
        lowball_app_mock_providers.application.auth_provider.get_client.assert_called_once_with(client_id_normal2)

    def test_failure_if_no_auth_provider(self,
                                         lowball_app_no_auth_provider,
                                         expected_no_auth_provider_error_response,
                                         admin1_token1_request_header,
                                         mock_lookup_token_filled,
                                         client_id_normal1):
        route = self.ROUTE.format(client_id_normal1)
        response = lowball_app_no_auth_provider.get(route, headers=admin1_token1_request_header)
        assert response.status_code == 503
        assert response.json["message"] == expected_no_auth_provider_error_response

    def test_success_outputs_for_client_data(self,
                                             lowball_app_mock_providers,
                                             mock_lookup_token_filled,
                                             mock_get_client_filled,
                                             get_client_request_response_rounds
                                             ):
        header, client_id, expected_response = get_client_request_response_rounds
        route = self.ROUTE.format(client_id)

        response = lowball_app_mock_providers.get(route, headers=header)
        print(response.json)
        assert response.status_code == 200
        assert response.json == expected_response
        assert call(client_id) in lowball_app_mock_providers.application.auth_provider.get_client.mock_calls


class TestUpdateClient:

    ROUTE = "/builtins/auth/clients/{}"

    def test_fails_with_no_auth_database(self, lowball_app_no_db, expected_no_auth_db_response, basic_header):
        route = self.ROUTE.format("anyclientid")
        response = lowball_app_no_db.post(route, headers=basic_header)
        assert response.status_code == 503
        assert response.json["message"] == expected_no_auth_db_response

    # assert authenticated
    def test_attempt_fails_when_not_authenticated_or_token_expired(self, lowball_app_mock_providers,
                                                                   not_authenticated_or_token_expired_header,
                                                                   client_id_normal1):
        route = self.ROUTE.format(client_id_normal1)
        response = lowball_app_mock_providers.post(route, headers=not_authenticated_or_token_expired_header,
                                                    json={})
        assert response.status_code != 200
        assert any([
            "Token Has Expired" in response.json["message"],
            "No token provided" in response.json["message"],
            "Token Is Invalid" in response.json["message"]
        ])

    # assert authenticated admin
    def test_attempt_fails_when_authenticated_but_not_admin(self, lowball_app_mock_providers,
                                                            non_admin_headers, client_id_normal1):
        route = self.ROUTE.format(client_id_normal1)
        response = lowball_app_mock_providers.post(route, headers=non_admin_headers,
                                                    json={})
        assert response.status_code == 401
        assert "Unauthorized" in response.json["message"]

    # fail no auth provider
    def test_failure_when_no_auth_provider_initialized(self, lowball_app_no_auth_provider,
                                                       expected_no_auth_provider_error_response,
                                                       admin1_token1_request_header,
                                                       mock_lookup_token_filled,
                                                       client_id_normal1):
        route = self.ROUTE.format(client_id_normal1)
        response = lowball_app_no_auth_provider.post(route, headers=admin1_token1_request_header)
        assert response.status_code == 503
        assert response.json["message"] == expected_no_auth_provider_error_response

    # fail no client
    def test_attempt_fails_when_client_does_not_exist(self,
                                                      lowball_app_mock_providers,
                                                      admin1_token1_request_header,
                                                      client_id_normal2,
                                                      mock_lookup_token_filled,
                                                      mock_get_client_filled
                                                      ):
        route = self.ROUTE.format(client_id_normal2)
        response = lowball_app_mock_providers.post(route, headers=admin1_token1_request_header, json={})
        print(response.json)
        assert response.status_code == 404
        lowball_app_mock_providers.application.auth_provider.get_client.assert_called_once_with(client_id_normal2)

    def test_attempt_fails_when_no_update_package_defined(self,
                                                          lowball_app_mock_providers,
                                                          admin1_token1_request_header,
                                                          client_id_normal1,
                                                          mock_lookup_token_filled,
                                                          mock_get_client_filled
                                                          ):
        route = self.ROUTE.format(client_id_normal1)
        response = lowball_app_mock_providers.post(route, headers=admin1_token1_request_header, json={})
        print(response.json)
        assert response.status_code == 501
        lowball_app_mock_providers.application.auth_provider.get_client.assert_called_once_with(client_id_normal1)
        assert "update_client_package_class" in response.json["message"]

    def test_attempt_fails_when_invalid_data_input(self,
                                                   lowball_app_mock_providers,
                                                   admin1_token1_request_header,
                                                   mock_lookup_token_filled,
                                                   mock_get_client_filled,
                                                   client_id_normal1,
                                                   expected_bad_package_data_error_response,
                                                   does_not_match_update_package_request,
                                                   mock_update_client_class
                                                   ):
        route = self.ROUTE.format(client_id_normal1)

        response = lowball_app_mock_providers.post(route,
                                                    headers=admin1_token1_request_header,
                                                    json=does_not_match_update_package_request)
        lowball_app_mock_providers.application.auth_provider.get_client.assert_called_once_with(client_id_normal1)
        assert response.status_code == 400
        assert response.json["message"] == expected_bad_package_data_error_response

    # success flow multiple output
    def test_attempt_successful_and_expected_output(self,
                                                    lowball_app_mock_providers,
                                                    admin1_token1_request_header,
                                                    mock_lookup_token_filled,
                                                    mock_get_client_filled,
                                                    client_id_normal1,
                                                    mock_self_update_class,
                                                    basic_client_data,
                                                    mock_update_client_rounds,
                                                    success_request,
                                                    mock_update_client_class
                                                    ):
        route = self.ROUTE.format(client_id_normal1)

        response = lowball_app_mock_providers.post(route,
                                                    json=success_request,
                                                    headers=admin1_token1_request_header)
        print(response.json)
        assert response.status_code == 200
        assert response.json == basic_client_data.to_dict()
        lowball_app_mock_providers.application.auth_provider.get_client.assert_called_once_with(client_id_normal1)
        lowball_app_mock_providers.application.auth_provider.update_client.assert_called_once()


class TestDeleteClient:
    ROUTE = "/builtins/auth/clients/{}"

    def test_fails_with_no_auth_database(self, lowball_app_no_db, expected_no_auth_db_response, basic_header):
        route = self.ROUTE.format("anyclientid")
        response = lowball_app_no_db.delete(route, headers=basic_header)
        assert response.status_code == 503
        assert response.json["message"] == expected_no_auth_db_response

    # assert authenticated
    def test_attempt_fails_when_not_authenticated_or_token_expired(self, lowball_app_mock_providers,
                                                                   not_authenticated_or_token_expired_header,
                                                                   client_id_normal1):
        route = self.ROUTE.format(client_id_normal1)

        response = lowball_app_mock_providers.delete(route, headers=not_authenticated_or_token_expired_header,
                                                     json={})
        assert response.status_code != 200
        assert any([
            "Token Has Expired" in response.json["message"],
            "No token provided" in response.json["message"],
            "Token Is Invalid" in response.json["message"]
        ])

    # assert authenticated admin
    def test_attempt_fails_when_authenticated_but_not_admin(self, lowball_app_mock_providers,
                                                            non_admin_headers, client_id_normal1):
        route = self.ROUTE.format(client_id_normal1)
        response = lowball_app_mock_providers.delete(route, headers=non_admin_headers,
                                                     json={})
        assert response.status_code == 401
        assert "Unauthorized" in response.json["message"]

    def test_failure_when_no_auth_provider_initialized(self, lowball_app_no_auth_provider,
                                                       expected_no_auth_provider_error_response,
                                                       admin1_token1_request_header,
                                                       mock_lookup_token_filled):
        response = lowball_app_no_auth_provider.delete(self.ROUTE, headers=admin1_token1_request_header)
        assert response.status_code == 503
        assert response.json["message"] == expected_no_auth_provider_error_response

    # fail no client
    def test_not_found_when_not_client(self,
                                       lowball_app_mock_providers,
                                       admin1_token1_request_header,
                                       client_id_normal2,
                                       mock_lookup_token_filled,
                                       mock_get_client_filled
                                       ):
        route = self.ROUTE.format(client_id_normal2)
        response = lowball_app_mock_providers.delete(route, headers=admin1_token1_request_header, json={})
        print(response.json)
        assert response.status_code == 404
        lowball_app_mock_providers.application.auth_provider.get_client.assert_called_once_with(client_id_normal2)

    # success route
    def test_calls_delete_successfully(self,
                                       lowball_app_mock_providers,
                                       admin1_token1_request_header,
                                       client_id_normal1,
                                       mock_lookup_token_filled,
                                       mock_get_client_filled,
                                       mock_delete_client_simple
                                       ):
        route = self.ROUTE.format(client_id_normal1)
        response = lowball_app_mock_providers.delete(route, headers=admin1_token1_request_header, json={})
        assert response.status_code == 204
        lowball_app_mock_providers.application.auth_provider.get_client.assert_called_once_with(client_id_normal1)
        lowball_app_mock_providers.application.auth_provider.delete_client.assert_called_once_with(client_id_normal1)


class TestEnableClient:
    ROUTE = "/builtins/auth/clients/{}/enable"

    def test_fails_with_no_auth_database(self, lowball_app_no_db, expected_no_auth_db_response, basic_header):
        route = self.ROUTE.format("anyclientid")
        response = lowball_app_no_db.post(route, headers=basic_header)
        assert response.status_code == 503
        assert response.json["message"] == expected_no_auth_db_response

    # assert authenticated
    def test_attempt_fails_when_not_authenticated_or_token_expired(self,
                                                                   lowball_app_mock_providers,
                                                                   not_authenticated_or_token_expired_header,
                                                                   client_id_normal1):
        route = self.ROUTE.format(client_id_normal1)
        response = lowball_app_mock_providers.post(route, headers=not_authenticated_or_token_expired_header,
                                                   json={})
        assert response.status_code != 200
        assert any([
            "Token Has Expired" in response.json["message"],
            "No token provided" in response.json["message"],
            "Token Is Invalid" in response.json["message"]
        ])

    # assert authenticated admin
    def test_attempt_fails_when_authenticated_but_not_admin(self, lowball_app_mock_providers,
                                                            non_admin_headers, client_id_normal1):
        route = self.ROUTE.format(client_id_normal1)
        response = lowball_app_mock_providers.post(route, headers=non_admin_headers,
                                                   json={})
        assert response.status_code == 401
        assert "Unauthorized" in response.json["message"]

    def test_failure_when_no_auth_provider_initialized(self,
                                                       lowball_app_no_auth_provider,
                                                       expected_no_auth_provider_error_response,
                                                       admin1_token1_request_header,
                                                       mock_lookup_token_filled):
        response = lowball_app_no_auth_provider.post(self.ROUTE, headers=admin1_token1_request_header)
        assert response.status_code == 503
        assert response.json["message"] == expected_no_auth_provider_error_response

    # fail no client
    def test_not_found_when_not_client(self,
                                       lowball_app_mock_providers,
                                       admin1_token1_request_header,
                                       client_id_normal2,
                                       mock_lookup_token_filled,
                                       mock_get_client_filled
                                       ):
        route = self.ROUTE.format(client_id_normal2)
        response = lowball_app_mock_providers.post(route, headers=admin1_token1_request_header, json={})
        print(response.json)
        assert response.status_code == 404
        lowball_app_mock_providers.application.auth_provider.get_client.assert_called_once_with(client_id_normal2)

    # success route
    def test_calls_enable_successfully(self,
                                       lowball_app_mock_providers,
                                       admin1_token1_request_header,
                                       client_id_normal1,
                                       mock_lookup_token_filled,
                                       mock_get_client_filled,
                                       mock_enable_client_simple
                                       ):
        route = self.ROUTE.format(client_id_normal1)
        response = lowball_app_mock_providers.post(route, headers=admin1_token1_request_header, json={})
        print(response.json)
        assert response.status_code == 204
        lowball_app_mock_providers.application.auth_provider.get_client.assert_called_once_with(client_id_normal1)
        lowball_app_mock_providers.application.auth_provider.enable_client.assert_called_once_with(client_id_normal1)

class TestDisableClient:

    ROUTE = "/builtins/auth/clients/{}/disable"

    def test_fails_with_no_auth_database(self, lowball_app_no_db, expected_no_auth_db_response, basic_header):
        route = self.ROUTE.format("anyclientid")
        response = lowball_app_no_db.post(route, headers=basic_header)
        assert response.status_code == 503
        assert response.json["message"] == expected_no_auth_db_response

    # assert authenticated
    def test_attempt_fails_when_not_authenticated_or_token_expired(self, lowball_app_mock_providers,
                                                                   not_authenticated_or_token_expired_header,
                                                                   client_id_normal1):
        route = self.ROUTE.format(client_id_normal1)

        response = lowball_app_mock_providers.post(route, headers=not_authenticated_or_token_expired_header,
                                                   json={})
        assert response.status_code != 200
        assert any([
            "Token Has Expired" in response.json["message"],
            "No token provided" in response.json["message"],
            "Token Is Invalid" in response.json["message"]
        ])

    # assert authenticated admin
    def test_attempt_fails_when_authenticated_but_not_admin(self, lowball_app_mock_providers,
                                                            non_admin_headers, client_id_normal1):
        route = self.ROUTE.format(client_id_normal1)
        response = lowball_app_mock_providers.post(route, headers=non_admin_headers,
                                                   json={})
        assert response.status_code == 401
        assert "Unauthorized" in response.json["message"]

    def test_failure_when_no_auth_provider_initialized(self,
                                                       lowball_app_no_auth_provider,
                                                       expected_no_auth_provider_error_response,
                                                       admin1_token1_request_header,
                                                       mock_lookup_token_filled):
        response = lowball_app_no_auth_provider.post(self.ROUTE, headers=admin1_token1_request_header)
        assert response.status_code == 503
        assert response.json["message"] == expected_no_auth_provider_error_response

    # fail no client
    def test_not_found_when_not_client(self,
                                       lowball_app_mock_providers,
                                       admin1_token1_request_header,
                                       client_id_normal2,
                                       mock_lookup_token_filled,
                                       mock_get_client_filled
                                       ):
        route = self.ROUTE.format(client_id_normal2)
        response = lowball_app_mock_providers.post(route, headers=admin1_token1_request_header, json={})
        print(response.json)
        assert response.status_code == 404
        lowball_app_mock_providers.application.auth_provider.get_client.assert_called_once_with(client_id_normal2)

    # success route
    def test_calls_disable_successfully(self,
                                        lowball_app_mock_providers,
                                        admin1_token1_request_header,
                                        client_id_normal1,
                                        mock_lookup_token_filled,
                                        mock_get_client_filled,
                                        mock_disable_client_simple
                                        ):
        route = self.ROUTE.format(client_id_normal1)
        response = lowball_app_mock_providers.post(route, headers=admin1_token1_request_header, json={})
        print(response.json)
        assert response.status_code == 204
        lowball_app_mock_providers.application.auth_provider.get_client.assert_called_once_with(client_id_normal1)
        lowball_app_mock_providers.application.auth_provider.disable_client.assert_called_once_with(
            client_id_normal1)


class TestRemoveClientRoles:

    ROUTE = "/builtins/auth/clients/{}/roles"

    def test_fails_with_no_auth_database(self, lowball_app_no_db, expected_no_auth_db_response, basic_header):
        route = self.ROUTE.format("anyclientid")
        response = lowball_app_no_db.delete(route, headers=basic_header)
        assert response.status_code == 503
        assert response.json["message"] == expected_no_auth_db_response

    # assert authenticated
    def test_attempt_fails_when_not_authenticated_or_token_expired(self, lowball_app_mock_providers,
                                                                   not_authenticated_or_token_expired_header,
                                                                   client_id_normal1):
        route = self.ROUTE.format(client_id_normal1)

        response = lowball_app_mock_providers.delete(route, headers=not_authenticated_or_token_expired_header,
                                                   json={})
        assert response.status_code != 200
        assert any([
            "Token Has Expired" in response.json["message"],
            "No token provided" in response.json["message"],
            "Token Is Invalid" in response.json["message"]
        ])

    # assert authenticated admin
    def test_attempt_fails_when_authenticated_but_not_admin(self, lowball_app_mock_providers,
                                                            non_admin_headers, client_id_normal1):
        route = self.ROUTE.format(client_id_normal1)
        response = lowball_app_mock_providers.delete(route, headers=non_admin_headers,
                                                   json={})
        assert response.status_code == 401
        assert "Unauthorized" in response.json["message"]

    def test_failure_when_no_auth_provider_initialized(self,
                                                       lowball_app_no_auth_provider,
                                                       expected_no_auth_provider_error_response,
                                                       admin1_token1_request_header,
                                                       mock_lookup_token_filled,
                                                       client_id_normal1):
        route = self.ROUTE.format(client_id_normal1)

        response = lowball_app_no_auth_provider.delete(route, headers=admin1_token1_request_header)
        assert response.status_code == 503
        assert response.json["message"] == expected_no_auth_provider_error_response

    # fail no client
    def test_not_found_when_not_client(self,
                                       lowball_app_mock_providers,
                                       admin1_token1_request_header,
                                       client_id_normal2,
                                       mock_lookup_token_filled,
                                       mock_get_client_filled
                                       ):
        route = self.ROUTE.format(client_id_normal2)
        response = lowball_app_mock_providers.delete(route, headers=admin1_token1_request_header, json={})
        print(response.json)
        assert response.status_code == 404
        lowball_app_mock_providers.application.auth_provider.get_client.assert_called_once_with(client_id_normal2)

    def test_calls_delete_roles_with_client_roles(self,
                                                  lowball_app_mock_providers,
                                                  admin1_token1_request_header,
                                                  client_id_normal1,
                                                  client_data_normal1,
                                                  mock_get_client_filled,
                                                  mock_lookup_token_filled,
                                                  mock_delete_roles
                                                  ):

        route = self.ROUTE.format(client_id_normal1)
        response = lowball_app_mock_providers.delete(route, headers=admin1_token1_request_header)

        assert response.status_code == 204
        lowball_app_mock_providers.application.auth_provider.get_client.assert_called_once_with(client_id_normal1)
        lowball_app_mock_providers.application.auth_provider.delete_roles.assert_called_once_with(client_id_normal1, client_data_normal1.roles)
    
    
class TestRemoveClientRole:
    
    ROUTE = "/builtins/auth/clients/{}/roles/{}"

    def test_fails_with_no_auth_database(self, lowball_app_no_db, expected_no_auth_db_response, basic_header):
        route = self.ROUTE.format("anyclientid", "anyrole")
        response = lowball_app_no_db.delete(route, headers=basic_header)
        assert response.status_code == 503
        assert response.json["message"] == expected_no_auth_db_response

    # assert authenticated
    def test_attempt_fails_when_not_authenticated_or_token_expired(self, lowball_app_mock_providers,
                                                                   not_authenticated_or_token_expired_header,
                                                                   client_id_normal1):
        route = self.ROUTE.format(client_id_normal1, "anyrole")

        response = lowball_app_mock_providers.delete(route, headers=not_authenticated_or_token_expired_header,
                                                     json={})
        assert response.status_code != 200
        assert any([
            "Token Has Expired" in response.json["message"],
            "No token provided" in response.json["message"],
            "Token Is Invalid" in response.json["message"]
        ])

    # assert authenticated admin
    def test_attempt_fails_when_authenticated_but_not_admin(self, lowball_app_mock_providers,
                                                            non_admin_headers, client_id_normal1):
        route = self.ROUTE.format(client_id_normal1, "anyrole")
        response = lowball_app_mock_providers.delete(route, headers=non_admin_headers,
                                                     json={})
        assert response.status_code == 401
        assert "Unauthorized" in response.json["message"]

    def test_failure_when_no_auth_provider_initialized(self,
                                                       lowball_app_no_auth_provider,
                                                       expected_no_auth_provider_error_response,
                                                       admin1_token1_request_header,
                                                       mock_lookup_token_filled,
                                                       client_id_normal1):
        route = self.ROUTE.format(client_id_normal1, "anyrole")

        response = lowball_app_no_auth_provider.delete(route, headers=admin1_token1_request_header)
        assert response.status_code == 503
        assert response.json["message"] == expected_no_auth_provider_error_response

    # fail no client
    def test_not_found_when_not_client(self,
                                       lowball_app_mock_providers,
                                       admin1_token1_request_header,
                                       client_id_normal2,
                                       mock_lookup_token_filled,
                                       mock_get_client_filled
                                       ):
        route = self.ROUTE.format(client_id_normal2, "anyrole")
        response = lowball_app_mock_providers.delete(route, headers=admin1_token1_request_header, json={})
        print(response.json)
        assert response.status_code == 404
        lowball_app_mock_providers.application.auth_provider.get_client.assert_called_once_with(client_id_normal2)
    
    def test_calls_delete_roles_with_role(self,
                                          lowball_app_mock_providers,
                                          admin1_token1_request_header,
                                          client_id_normal1,
                                          mock_lookup_token_filled,
                                          mock_get_client_filled,
                                          base_role1,
                                          mock_delete_roles
                                          ):
        role = base_role1[0]
        route = self.ROUTE.format(client_id_normal1, role)
        response = lowball_app_mock_providers.delete(route, headers=admin1_token1_request_header)
        
        assert response.status_code == 204
        lowball_app_mock_providers.application.auth_provider.get_client.assert_called_once_with(client_id_normal1)
        lowball_app_mock_providers.application.auth_provider.delete_roles.assert_called_once_with(client_id_normal1, base_role1)


class TestAddClientRole:
    ROUTE = "/builtins/auth/clients/{}/roles/{}"

    def test_fails_with_no_auth_database(self, lowball_app_no_db, expected_no_auth_db_response, basic_header):
        route = self.ROUTE.format("anyclientid", "anyrole")
        response = lowball_app_no_db.post(route, headers=basic_header)
        assert response.status_code == 503
        assert response.json["message"] == expected_no_auth_db_response

    # assert authenticated
    def test_attempt_fails_when_not_authenticated_or_token_expired(self, lowball_app_mock_providers,
                                                                   not_authenticated_or_token_expired_header,
                                                                   client_id_normal1):
        route = self.ROUTE.format(client_id_normal1, "anyrole")

        response = lowball_app_mock_providers.post(route, headers=not_authenticated_or_token_expired_header,
                                                     json={})
        assert response.status_code != 200
        assert any([
            "Token Has Expired" in response.json["message"],
            "No token provided" in response.json["message"],
            "Token Is Invalid" in response.json["message"]
        ])

    # assert authenticated admin
    def test_attempt_fails_when_authenticated_but_not_admin(self, lowball_app_mock_providers,
                                                            non_admin_headers, client_id_normal1):
        route = self.ROUTE.format(client_id_normal1, "anyrole")
        response = lowball_app_mock_providers.post(route, headers=non_admin_headers,
                                                     json={})
        assert response.status_code == 401
        assert "Unauthorized" in response.json["message"]

    def test_failure_when_no_auth_provider_initialized(self,
                                                       lowball_app_no_auth_provider,
                                                       expected_no_auth_provider_error_response,
                                                       admin1_token1_request_header,
                                                       mock_lookup_token_filled,
                                                       client_id_normal1):
        route = self.ROUTE.format(client_id_normal1, "anyrole")

        response = lowball_app_no_auth_provider.post(route, headers=admin1_token1_request_header)
        assert response.status_code == 503
        assert response.json["message"] == expected_no_auth_provider_error_response

    # fail no client
    def test_not_found_when_not_client(self,
                                       lowball_app_mock_providers,
                                       admin1_token1_request_header,
                                       client_id_normal2,
                                       mock_lookup_token_filled,
                                       mock_get_client_filled
                                       ):
        route = self.ROUTE.format(client_id_normal2, "anyrole")
        response = lowball_app_mock_providers.post(route, headers=admin1_token1_request_header, json={})
        print(response.json)
        assert response.status_code == 404
        lowball_app_mock_providers.application.auth_provider.get_client.assert_called_once_with(client_id_normal2)

    def test_calls_delete_roles_with_role(self,
                                          lowball_app_mock_providers,
                                          admin1_token1_request_header,
                                          client_id_normal1,
                                          mock_lookup_token_filled,
                                          mock_get_client_filled,
                                          base_role1,
                                          mock_add_roles
                                          ):
        role = base_role1[0]
        route = self.ROUTE.format(client_id_normal1, role)
        response = lowball_app_mock_providers.post(route, headers=admin1_token1_request_header)

        assert response.status_code == 204
        lowball_app_mock_providers.application.auth_provider.get_client.assert_called_once_with(client_id_normal1)
        lowball_app_mock_providers.application.auth_provider.add_roles.assert_called_once_with(client_id_normal1,
                                                                                                  base_role1)

